from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.manybody.record_db_models import ManybodyRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.auth import UserInfo
from qcportal.manybody import ManybodySpecification, compare_manybody_records
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import SinglepointProtocols, QCSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import (
    compare_manybody_specs,
    test_specs,
    load_procedure_data,
    generate_task_key,
    load_record_data,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal.managers import ManagerName
    from sqlalchemy.orm.session import Session


@pytest.mark.parametrize("spec", test_specs)
def test_manybody_socket_add_get(storage_socket: SQLAlchemySocket, session: Session, spec: ManybodySpecification):
    water2 = load_molecule_data("water_dimer_minima")
    water4 = load_molecule_data("water_stacked")

    time_0 = now_at_utc()
    meta, ids = storage_socket.records.manybody.add([water2, water4], spec, "tag1", PriorityEnum.low, None, True)
    time_1 = now_at_utc()
    assert meta.success

    recs = [session.get(ManybodyRecordORM, i) for i in ids]

    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "manybody"
        assert r.status == RecordStatusEnum.waiting
        assert compare_manybody_specs(spec, r.specification.model_dict())

        # Service queue entry should exist with the proper tag and priority
        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert recs[0].initial_molecule.identifiers["molecule_hash"] == water2.get_hash()
    assert recs[1].initial_molecule.identifiers["molecule_hash"] == water4.get_hash()


def test_manybody_socket_add_same_1(storage_socket: SQLAlchemySocket):
    spec = ManybodySpecification(
        program="qcmanybody",
        levels={
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    water2 = load_molecule_data("water_dimer_minima")
    water4 = load_molecule_data("water_stacked")

    meta, id1 = storage_socket.records.manybody.add([water2, water4], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 2
    assert meta.inserted_idx == [0, 1]

    meta, id2 = storage_socket.records.manybody.add([water4, water2], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 2
    assert meta.existing_idx == [0, 1]
    assert id1[0] == id2[1]
    assert id1[1] == id2[0]


@pytest.mark.parametrize(
    "test_data_name",
    [
        "mb_cp_he4_psi4_mp2",
        "mb_all_he4_psi4_multi",
        "mb_all_he4_psi4_multiss",
    ],
)
def test_manybody_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName, test_data_name: str
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data(test_data_name)

    storage_socket.users.add(UserInfo(username="submit_user", role="submit", enabled=True))

    meta_1, id_1 = storage_socket.records.manybody.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, "submit_user", True
    )
    id_1 = id_1[0]
    assert meta_1.success

    time_0 = now_at_utc()
    finished, n_singlepoints = run_service(
        storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 100
    )
    time_1 = now_at_utc()

    assert finished is True

    rec = session.get(ManybodyRecordORM, id_1)

    assert rec.status == RecordStatusEnum.complete
    assert time_0 < rec.modified_on < time_1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 1
    assert rec.compute_history[-1].status == RecordStatusEnum.complete
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is None

    desc_info = storage_socket.records.get_short_descriptions([id_1])[0]
    short_desc = desc_info["description"]
    assert desc_info["record_type"] == rec.record_type
    assert desc_info["created_on"] == rec.created_on
    assert rec.specification.program in short_desc

    out = rec.compute_history[-1].outputs["stdout"].get_output()
    assert "All manybody singlepoint computations are complete" in out

    unique_sp = set(x.singlepoint_id for x in rec.clusters)
    assert len(unique_sp) == n_singlepoints


def test_manybody_socket_run_duplicate(
    storage_socket: SQLAlchemySocket,
    session: Session,
    activated_manager_name: ManagerName,
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data("mb_cp_he4_psi4_mp2")

    meta_1, id_1 = storage_socket.records.manybody.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, True
    )
    id_1 = id_1[0]
    assert meta_1.success

    run_service(storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 1000)

    rec_1 = session.get(ManybodyRecordORM, id_1)
    assert rec_1.status == RecordStatusEnum.complete
    sp_ids_1 = [x.singlepoint_id for x in rec_1.clusters]

    # Submit again, without duplicate checking
    meta_2, id_2 = storage_socket.records.manybody.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, False
    )
    id_2 = id_2[0]
    assert meta_2.success
    assert id_2 != id_1

    run_service(storage_socket, activated_manager_name, id_2, generate_task_key, result_data_1, 1000)

    rec_2 = session.get(ManybodyRecordORM, id_2)
    assert rec_2.status == RecordStatusEnum.complete
    sp_ids_2 = [x.singlepoint_id for x in rec_2.clusters]

    assert set(sp_ids_1).isdisjoint(sp_ids_2)


def test_manybody_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "mb_all_he4_psi4_multi",
        "mb_all_he4_psi4_multiss",
        "mb_cp_he4_psi4_mp2",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_manybodys(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)
        compare_manybody_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"
        for cl in rec_1.cluster_records_:
            assert cl.singlepoint_record.creator_user == "submit_user"



"""
Main/Root socket for accessing the database
"""

from __future__ import annotations

import importlib
import logging
import os
import shutil
from contextlib import contextmanager
from functools import lru_cache
from typing import TYPE_CHECKING

from sqlalchemy import create_engine, exc, event, inspect, select, union, MetaData, Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

import qcfractal

if TYPE_CHECKING:
    from typing import List, Optional, Generator, Any
    from sqlalchemy.orm.session import Session
    from ..config import FractalConfig, DatabaseConfig


class SQLAlchemySocket:
    """
    Main/Root socket accessing/managing an SQLAlchemy database
    """

    def __init__(self, qcf_config: FractalConfig):
        self.qcf_config = qcf_config

        # By default, disable watching
        self._finished_queue = None

        # Logging data
        self.logger = logging.getLogger("SQLAlchemySocket")

        self.logger.info(f"SQLAlchemy attempt to connect to {qcf_config.database.safe_uri}.")

        # Connect to DB and create session
        # If the pool size given in the config is zero, then that corresponds to using a NullPool here.
        # Note that this is different than SQLAlchemy, where pool_size = 0 means unlimited
        # If pool_size in the config is non-zero, then set the pool class to None (meaning use
        # SQLAlchemy default)
        if qcf_config.database.pool_size == 0:
            self.engine = create_engine(
                self.qcf_config.database.sqlalchemy_url,
                echo=qcf_config.database.echo_sql,
                poolclass=NullPool,
                future=True,
            )
        else:
            self.engine = create_engine(
                self.qcf_config.database.sqlalchemy_url,
                echo=qcf_config.database.echo_sql,
                pool_size=qcf_config.database.pool_size,
                future=True,
            )

        self.logger.info(
            "Connected SQLAlchemy to DB dialect {} with driver {}".format(self.engine.dialect.name, self.engine.driver)
        )

        # Create an equivalent engine with autocommit
        # See https://docs.sqlalchemy.org/en/20/orm/session_transaction.html#session-transaction-isolation
        self.autocommit_engine = self.engine.execution_options(isolation_level="AUTOCOMMIT")

        # Handle multiprocessing w/ sqlalchemy
        # https://docs.sqlalchemy.org/en/14/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork
        @event.listens_for(self.engine, "connect")
        def connect(dbapi_connection, connection_record):
            connection_record.info["pid"] = os.getpid()

        @event.listens_for(self.engine, "checkout")
        def checkout(dbapi_connection, connection_record, connection_proxy):
            pid = os.getpid()
            if connection_record.info["pid"] != pid:
                connection_record.dbapi_connection = connection_proxy.dbapi_connection = None
                raise exc.DisconnectionError(
                    "Connection record belongs to pid %s, "
                    "attempting to check out in pid %s" % (connection_record.info["pid"], pid)
                )

        # Check to see if the db is up-to-date
        self._check_db_revision(self.engine)

        self.Session = sessionmaker(bind=self.engine, future=True)

        # Create/initialize the subsockets
        from ..components.internal_jobs.socket import InternalJobSocket
        from ..components.molecules.socket import MoleculeSocket
        from ..components.auth.user_socket import UserSocket
        from ..components.auth.group_socket import GroupSocket
        from ..components.auth.auth_socket import AuthSocket
        from ..components.serverinfo.socket import ServerInfoSocket
        from ..components.managers.socket import ManagerSocket
        from ..components.tasks.socket import TaskSocket
        from ..components.services.socket import ServiceSocket
        from ..components.external_files import ExternalFileSocket
        from ..components.record_socket import RecordSocket
        from ..components.dataset_socket import DatasetSocket
        from ..components.project_socket import ProjectSocket

        # Internal job socket goes first - others may depend on this
        self.internal_jobs = InternalJobSocket(self)

        # Then the rest
        self.serverinfo = ServerInfoSocket(self)
        self.molecules = MoleculeSocket(self)
        self.records = RecordSocket(self)
        self.datasets = DatasetSocket(self)
        self.projects = ProjectSocket(self)  # Must be done after dataset and record sockets
        self.external_files = ExternalFileSocket(self)
        self.tasks = TaskSocket(self)
        self.services = ServiceSocket(self)
        self.managers = ManagerSocket(self)
        self.users = UserSocket(self)
        self.groups = GroupSocket(self)
        self.auth = AuthSocket(self)

    def __str__(self) -> str:
        return f"<SQLAlchemySocket: address='{self.qcf_config.database.safe_uri}`>"

    @staticmethod
    def alembic_commands(db_config: DatabaseConfig) -> List[str]:
        """
        Get the components of an alembic command that can be used on the command line

        This will find the almembic command and also add the uri and alembic configuration information
        to the command line.

        Returns
        -------
        List[str]
            Components of an alembic command line as a list of strings
        """

        # Find the path to the almebic ini
        alembic_ini = os.path.join(qcfractal.qcfractal_topdir, "alembic.ini")
        alembic_path = shutil.which("alembic")

        if alembic_path is None:
            raise RuntimeError("Cannot find the 'alembic' command. Is it installed?")
        return [
            alembic_path,
            "-c",
            alembic_ini,
            "-x",
            "uri=" + db_config.database_uri,
        ]

    @staticmethod
    def get_alembic_config(db_config: DatabaseConfig):
        """
        Obtain an alembic Config object given a QCFractal database configuration
        """

        from alembic.config import Config

        alembic_ini = os.path.join(qcfractal.qcfractal_topdir, "alembic.ini")
        alembic_cfg = Config(alembic_ini)

        # Tell alembic to not set up logging. We already did that
        alembic_cfg.set_main_option("skip_logging", "True")
        alembic_cfg.set_main_option("sqlalchemy.url", db_config.database_uri)

        return alembic_cfg

    @staticmethod
    def create_database_tables(db_config: DatabaseConfig):
        """
        Create all the tables in a database

        The database is expected to exist already
        """

        logger = logging.getLogger("SQLAlchemySocket")

        # Register all classes that derive from the BaseORM
        importlib.import_module("qcfractal.components.register_all")

        # create the tables via sqlalchemy
        logger.info(f"Creating tables for database: {db_config.safe_uri}")
        engine = create_engine(db_config.sqlalchemy_url, echo=False, poolclass=NullPool, future=True)

        from qcfractal.db_socket.base_orm import BaseORM

        try:
            BaseORM.metadata.create_all(engine)
        except Exception as e:
            raise RuntimeError(f"SQLAlchemy Connection Error\n{str(e)}")

        SQLAlchemySocket.populate_defaults(engine)

        # update alembic_version table with the current version
        logger.debug(f"Stamping Database with current alembic revision")
        from alembic import command

        alembic_cfg = SQLAlchemySocket.get_alembic_config(db_config)
        command.stamp(alembic_cfg, "head")

    @staticmethod
    def populate_defaults(engine):
        """
        Populates any default values in the database
        """

        # Used to populate roles, but not anymore. Leaving for future needs
        pass

    @staticmethod
    def upgrade_database(db_config: DatabaseConfig, revision: str = "head") -> None:
        """
        Upgrade the database schema using the latest alembic revision.
        """

        from alembic import command

        alembic_cfg = SQLAlchemySocket.get_alembic_config(db_config)
        command.upgrade(alembic_cfg, revision)

    @staticmethod
    def _check_db_revision(engine: Engine):
        """
        Checks to make sure the database is up-to-date, given an engine

        Will raise an exception if it is not up-to-date
        """

        from alembic.migration import MigrationContext
        from alembic.script import ScriptDirectory

        script_dir = os.path.join(qcfractal.qcfractal_topdir, "alembic")
        script = ScriptDirectory(script_dir)
        heads = script.get_heads()

        conn = engine.connect()
        context = MigrationContext.configure(connection=conn)
        current_rev = context.get_current_revision()

        try:
            if len(heads) > 1:
                raise RuntimeError("Multiple alembic revision heads not supported")

            if heads[0] != current_rev:
                raise RuntimeError(
                    "Database needs migration. Please run `qcfractal-server upgrade-db` (after backing up!)"
                )
        finally:
            conn.close()

    @staticmethod
    def check_db_revision(db_config: DatabaseConfig):
        """
        Checks to make sure the database is up-to-date, given a configuration

        Will raise an exception if it is not up-to-date
        """

        engine = create_engine(db_config.sqlalchemy_url, echo=False, poolclass=NullPool, future=True)
        SQLAlchemySocket._check_db_revision(engine)

    @contextmanager
    def session_scope(self, read_only: bool = False):
        """Provide a session as a context manager

        You can use this as a context manager (ie, `with session_scope() as session`).
        After the `with` block exits, the session is committed (if not `read_only`)
        and then closed. On exception, the session is rolled back.
        """

        if read_only:
            session = self.Session(bind=self.autocommit_engine)
        else:
            session = self.Session(bind=self.engine)
        try:
            yield session

            if not read_only:
                session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @contextmanager
    def optional_session(self, existing_session: Optional[Session], read_only: bool = False):
        """Provide a session as a context manager

        Use the existing session if available, otherwise use a new session

        If an existing session is used, it is automatically flushed at the end, but not committed

        This is meant to be used with `with`, and where existing_session may be None.

         .. code-block:: python

            def somefunction(session: Optional[Session] = None):
                # will use session if not None, or will create a new session
                with storage_socket.optional_session(session) as s:
                    s.add(stuff)

        Parameters
        ----------
        existing_session
            An optional, existing sqlalchemy session
        """

        if existing_session is not None:
            yield existing_session

            if not read_only:
                existing_session.flush()

        else:
            with self.session_scope(read_only) as session:
                yield session

    def set_finished_watch(self, finished_queue):
        """
        Set the finished watch queue to the given multiprocessing queue

        When a calculation finishes, its record ID and status will be placed in this queue
        """

        self._finished_queue = finished_queue

    def notify_finished_watch(self, record_id, status):
        """
        Place information into the queue used to notify of finished calculations
        """

        if self._finished_queue is not None:
            # Don't want to block here. Just put it in the queue and move on
            self._finished_queue.put((int(record_id), status), block=False)

    @lru_cache()
    def find_referencing_cols(self, column):
        """
        Finds columns that have a foreign key referencing the given column

        For example, `find_referencing_cols(LargeBinaryORM.id)` will find all columns (in all tables in the database)
        that reference `LargeBinaryORM.id`
        """

        md = MetaData()
        md.reflect(bind=self.engine)

        insp_col = inspect(column)
        table = insp_col.table

        this_table_name = table.name
        this_column_name = insp_col.name

        db_reflection = inspect(self.engine)

        referencing_cols = []

        all_fkc = db_reflection.get_multi_foreign_keys()

        # Looking at all foreign keys, we want all that refer to this table and column
        # (constrained_columns is the columns on the other table)
        for (schema, table_name), fks in all_fkc.items():
            for fk in fks:
                if fk["referred_table"] == this_table_name and this_column_name in fk["referred_columns"]:
                    # Find what column refers to this one. It may be part of a composite fk
                    col_idx = fk["referred_columns"].index(this_column_name)
                    tb = md.tables[table_name]

                    # Turns it into something like BaseRecord.id (or the equivalent of a Table)
                    referencing_cols.append(getattr(tb.columns, fk["constrained_columns"][col_idx]))

        return referencing_cols

    def find_orphaned_rows(self, session: Session, column):
        """
        Finds any rows in a table that are not referenced elsewhere

        Only the given column is searched to find references
        """

        cte_parts = []
        referencing_columns = self.find_referencing_cols(column)
        for rc in referencing_columns:
            cp = select(rc.label("ref_col"))
            cte_parts.append(cp)

        cte = union(*cte_parts).cte()
        stmt = select(column)
        stmt = stmt.join(cte, cte.c.ref_col == column, isouter=True)
        stmt = stmt.where(cte.c.ref_col.is_(None))

        r = session.execute(stmt).scalars().all()
        return r



from flask import Blueprint

from qcfractal.flask_app.load_user import load_logged_in_user

api_v1 = Blueprint("api", __name__, url_prefix="/api/v1")


@api_v1.before_request
def load_user_info():
    load_logged_in_user()



from __future__ import annotations

from sqlalchemy import event, DDL

from qcfractal.components.record_db_models import OutputStoreORM

# Mark the storage of the data column as external
event.listen(
    OutputStoreORM.__table__,
    "after_create",
    DDL("ALTER TABLE output_store ALTER COLUMN data SET STORAGE EXTERNAL").execute_if(dialect=("postgresql")),
)



from __future__ import annotations

from sqlalchemy import Column, Integer, String, Enum, TIMESTAMP, BigInteger
from sqlalchemy.dialects.postgresql import JSONB

from qcfractal.db_socket.base_orm import BaseORM
from qcportal.external_files import ExternalFileStatusEnum, ExternalFileTypeEnum
from qcportal.utils import now_at_utc


class ExternalFileORM(BaseORM):
    """
    Table for storing molecules
    """

    __tablename__ = "external_file"

    id = Column(Integer, primary_key=True)
    file_type = Column(Enum(ExternalFileTypeEnum), nullable=False)

    created_on = Column(TIMESTAMP, default=now_at_utc, nullable=False)
    status = Column(Enum(ExternalFileStatusEnum), nullable=False)

    file_name = Column(String, nullable=False)
    description = Column(String, nullable=True)
    provenance = Column(JSONB, nullable=False)

    sha256sum = Column(String, nullable=False)
    file_size = Column(BigInteger, nullable=False)

    bucket = Column(String, nullable=False)
    object_key = Column(String, nullable=False)

    __mapper_args__ = {"polymorphic_on": "file_type"}

    _qcportal_model_excludes__ = ["object_key", "bucket"]



from sqlalchemy import union, select

from qcfractal.components.gridoptimization.dataset_db_models import (
    dataset_records_select as gridoptimization_dataset_records,
)
from qcfractal.components.manybody.dataset_db_models import dataset_records_select as manybody_dataset_records
from qcfractal.components.neb.dataset_db_models import dataset_records_select as neb_dataset_records
from qcfractal.components.optimization.dataset_db_models import dataset_records_select as optimization_dataset_records
from qcfractal.components.reaction.dataset_db_models import dataset_records_select as reaction_dataset_records
from qcfractal.components.record_db_views import RecordChildrenView
from qcfractal.components.singlepoint.dataset_db_models import dataset_records_select as singlepoint_dataset_records
from qcfractal.components.torsiondrive.dataset_db_models import dataset_records_select as torsiondrive_dataset_records
from qcfractal.db_socket.base_orm import BaseORM
from qcfractal.db_socket.db_views import view

DatasetDirectRecordsView = view(
    "dataset_direct_records_view",
    BaseORM.metadata,
    union(
        *singlepoint_dataset_records,
        *optimization_dataset_records,
        *gridoptimization_dataset_records,
        *torsiondrive_dataset_records,
        *manybody_dataset_records,
        *reaction_dataset_records,
        *neb_dataset_records,
    ),
)


DatasetRecordsView = view(
    "dataset_records_view",
    BaseORM.metadata,
    union(
        select(DatasetDirectRecordsView.c.dataset_id, DatasetDirectRecordsView.c.record_id),
        select(DatasetDirectRecordsView.c.dataset_id, RecordChildrenView.c.child_id.label("record_id")).join(
            RecordChildrenView, DatasetDirectRecordsView.c.record_id == RecordChildrenView.c.parent_id
        ),
    ),
)



"""
Classes for handling periodic tasks required by QCFractal

QCFractal requires some periodic maintenance. These include updating services, pruning
dead managers, and updating statistics.
"""

from __future__ import annotations

import multiprocessing
import threading
from typing import TYPE_CHECKING, Optional

from .db_socket.socket import SQLAlchemySocket

if TYPE_CHECKING:
    from .config import FractalConfig


class FractalJobRunner:
    """
    Runs internal QCFractal jobs

    QCFractal requires some jobs to be run periodically or otherwise in the background. This class will run these jobs.
    """

    def __init__(
        self,
        qcf_config: FractalConfig,
        finished_queue: Optional[multiprocessing.Queue] = None,
    ):
        """
        Parameters
        ----------
        qcf_config: FractalConfig
            Configuration for the QCFractal server
        """

        self.storage_socket = SQLAlchemySocket(qcf_config)
        self.storage_socket.set_finished_watch(finished_queue)
        self._end_event = threading.Event()

    def start(self) -> None:
        """
        Start running the periodic tasks in the foreground

        This function will block until interrupted
        """

        self.storage_socket.internal_jobs.run_loop(self._end_event)

    def stop(self) -> None:
        """
        Stop running the periodic tasks
        """

        self._end_event.set()



from __future__ import annotations

import importlib
import logging
import queue
from typing import TYPE_CHECKING

from flask import Flask, current_app
from flask_cors import CORS
from flask_jwt_extended import JWTManager
from werkzeug.local import LocalProxy
from werkzeug.routing import IntegerConverter

from .flask_session import QCFFlaskSessionInterface
from .flask_socket import FlaskStorageSocket
from .home_v1 import home_v1
from ..db_socket import SQLAlchemySocket

if TYPE_CHECKING:
    from ..config import FractalConfig
    from typing import Optional

app_storage_sockets = FlaskStorageSocket()


def _get_storage_socket() -> SQLAlchemySocket:
    return app_storage_sockets.get_socket(current_app._get_current_object())


storage_socket = LocalProxy(_get_storage_socket)

jwt = JWTManager()


# Some routes allow for negative integers (ie, list index)
# See https://github.com/pallets/flask/issues/2643
class SignedIntConverter(IntegerConverter):
    regex = r"-?\d+"


def create_flask_app(qcfractal_config: FractalConfig, finished_queue: Optional[queue.Queue] = None):
    app = Flask(__name__)

    app.url_map.converters["signed_int"] = SignedIntConverter

    app.logger = logging.getLogger("fractal_flask_app")
    app.logger.info(f"Creating flask app")

    # Read in and store the qcfractal configuration for later use
    app.config["QCFRACTAL_CONFIG"] = qcfractal_config

    # Configure the flask app

    # Some defaults (but can be overridden)
    # must be set to false to avoid restarting
    app.config["DEBUG"] = False

    # Never propagate exceptions. This uses the default error pages
    # which are HTML, but we are using json...
    app.config["PROPAGATE_EXCEPTIONS"] = False

    app.config["SECRET_KEY"] = qcfractal_config.api.secret_key
    app.config["JWT_SECRET_KEY"] = qcfractal_config.api.jwt_secret_key
    app.config["JWT_ACCESS_TOKEN_EXPIRES"] = qcfractal_config.api.jwt_access_token_expires
    app.config["JWT_REFRESH_TOKEN_EXPIRES"] = qcfractal_config.api.jwt_refresh_token_expires
    app.config["JWT_TOKEN_LOCATION"] = ["headers", "cookies"]
    app.config["SESSION_COOKIE_NAME"] = qcfractal_config.api.user_session_cookie_name
    app.config["PERMANENT_SESSION_LIFETIME"] = qcfractal_config.api.user_session_max_age

    # Where we store user-uploaded files for processing
    app.config["UPLOAD_FOLDER"] = qcfractal_config.upload_directory

    # Any additional configuration
    if qcfractal_config.api.extra_flask_options:
        app.config.update(**qcfractal_config.api.extra_flask_options)

    jwt.init_app(app)

    if qcfractal_config.cors.enabled:
        app.config["CORS_ORIGINS"] = qcfractal_config.cors.origins
        app.config["CORS_SUPPORTS_CREDENTIALS"] = qcfractal_config.cors.supports_credentials
        app.config["CORS_HEADERS"] = qcfractal_config.cors.headers
        CORS(app)

    # Initialize the database socket, API logger, and view handler
    app_storage_sockets.init_app(app, finished_queue=finished_queue)

    # Initialize the session interface after the storage socket
    app.session_interface = QCFFlaskSessionInterface(app)

    # Registers the various error and before/after request handlers
    importlib.import_module("qcfractal.flask_app.handlers")

    # Register all the routes in the other files.
    # Must be done before registering the blueprint
    importlib.import_module("qcfractal.flask_app.api_v1.routes")
    importlib.import_module("qcfractal.flask_app.auth_v1.routes")
    importlib.import_module("qcfractal.flask_app.compute_v1.routes")
    importlib.import_module("qcfractal.components.register_all")

    from .auth_v1.blueprint import auth_v1
    from .api_v1.blueprint import api_v1
    from .compute_v1.blueprint import compute_v1

    app.register_blueprint(home_v1)
    app.register_blueprint(api_v1)
    app.register_blueprint(auth_v1)
    app.register_blueprint(compute_v1)

    return app



"""add public id to sessions table

Revision ID: 8738641e0b90
Revises: b664ee54c62a
Create Date: 2025-04-05 20:19:36.318495

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8738641e0b90"
down_revision = "b664ee54c62a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("user_session", sa.Column("public_id", sa.Integer(), nullable=False))
    op.create_unique_constraint("ux_user_session_session_id", "user_session", ["session_id"])

    op.execute(sa.text("""create sequence user_session_public_id_seq"""))
    op.drop_constraint("user_session_pkey", "user_session")

    op.execute(
        sa.text(
            """ALTER TABLE user_session ALTER COLUMN public_id SET DEFAULT nextval('user_session_public_id_seq'::regclass)"""
        )
    )
    op.execute(sa.text("""ALTER SEQUENCE user_session_public_id_seq OWNED BY user_session.public_id"""))
    op.execute(sa.text("""CREATE UNIQUE INDEX user_session_pkey ON user_session USING btree (public_id)"""))
    op.execute(
        sa.text(
            """alter table user_session add constraint user_session_pkey PRIMARY KEY using index user_session_pkey"""
        )
    )

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Downgrade is not supported")



"""Remove unused columns from datasets

Revision ID: 236316a77635
Revises: d5988aa750ae
Create Date: 2025-03-13 15:38:25.315717

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "236316a77635"
down_revision = "d5988aa750ae"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("base_dataset", "group")
    op.drop_column("base_dataset", "extras")
    op.drop_column("base_dataset", "visibility")

    # Rename metadata to extras
    op.alter_column("base_dataset", "metadata", new_column_name="extras")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("base_dataset", "extras", new_column_name="metadata")
    op.add_column("base_dataset", sa.Column("visibility", sa.BOOLEAN(), autoincrement=False, nullable=False))
    op.add_column(
        "base_dataset", sa.Column("extras", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=False)
    )
    op.add_column("base_dataset", sa.Column("group", sa.VARCHAR(length=100), autoincrement=False, nullable=False))
    # ### end Alembic commands ###



from flask import current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.compute_v1.blueprint import compute_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.base_models import CommonBulkGetNamesBody
from qcportal.exceptions import LimitExceededError
from qcportal.managers import (
    ManagerActivationBody,
    ManagerUpdateBody,
    ManagerStatusEnum,
    ManagerQueryFilters,
    ManagerQueryAvailableFilters,
)
from qcportal.utils import calculate_limit


##################################################
# Routes that deal with manager activation, etc
##################################################


@compute_v1.route("/managers", methods=["POST"])
@wrap_global_route("managers", "add")
def activate_manager_v1(body_data: ManagerActivationBody):
    return storage_socket.managers.activate(
        name_data=body_data.name_data,
        manager_version=body_data.manager_version,
        username=body_data.username,
        programs=body_data.programs,
        compute_tags=body_data.compute_tags,
    )


@compute_v1.route("/managers/<string:name>", methods=["PATCH"])
@wrap_global_route("managers", "modify")
def update_manager_v1(name: str, body_data: ManagerUpdateBody):
    # This endpoint is used for heartbeats and deactivation

    # Will raise an exception if manager is not active
    storage_socket.managers.update_resource_stats(
        name=name,
        total_cpu_hours=body_data.total_cpu_hours,
        active_tasks=body_data.active_tasks,
        active_cores=body_data.active_cores,
        active_memory=body_data.active_memory,
    )

    # Deactivate if specified
    if body_data.status != ManagerStatusEnum.active:
        storage_socket.managers.deactivate([name])


######################################################
# Routes for the user API to get information about managers
######################################################


@api_v1.route("/managers/<string:name>", methods=["GET"])
@wrap_global_route("managers", "read")
def get_managers_v1(name: str):
    return storage_socket.managers.get([name])[0]


@api_v1.route("/managers/bulkGet", methods=["POST"])
@wrap_global_route("managers", "read")
def bulk_get_managers_v1(body_data: CommonBulkGetNamesBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_managers
    if len(body_data.names) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.names)} manager records - limit is {limit}")

    return storage_socket.managers.get(body_data.names, body_data.include, body_data.exclude, body_data.missing_ok)


@api_v1.route("/managers/<string:name>/log", methods=["GET"])
@wrap_global_route("managers", "read")
def get_manager_log_v1(name: str):
    return storage_socket.managers.get_log(name)


@api_v1.route("/managers/query", methods=["POST"])
@wrap_global_route("managers", "read")
def query_managers_v1(body_data: ManagerQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_managers
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.managers.query(body_data)


@api_v1.route("/managers/queryActive", methods=["POST"])
@wrap_global_route("managers", "read")
def query_active_managers_v1(body_data: ManagerQueryAvailableFilters):
    return storage_socket.managers.query_active(body_data.compute_tag, body_data.programs)



"""remove explicit fix_ columns in molecule

Revision ID: 49cfda6eb87e
Revises: de44b1b34097
Create Date: 2024-01-06 19:18:04.534992

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "49cfda6eb87e"
down_revision = "de44b1b34097"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("molecule", "fix_orientation")
    op.drop_column("molecule", "fix_com")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("molecule", sa.Column("fix_com", sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.add_column("molecule", sa.Column("fix_orientation", sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.execute(sa.text("UPDATE molecule SET fix_com = True, fix_orientation = True"))
    op.alter_column("molecule", "fix_com", nullable=False)
    op.alter_column("molecule", "fix_orientation", nullable=False)
    # ### end Alembic commands ###



from functools import partial
from typing import Any, Optional

import pytest

from qcarchivetesting import test_users
from qcfractal.testing_helpers import mname1
from qcportal import PortalClient
from qcportal import PortalRequestError
from qcportal.auth import UserInfo, GroupInfo
from qcportal.managers import ManagerActivationBody, ManagerUpdateBody
from qcportal.molecules import Molecule
from qcportal.singlepoint import SinglepointInput
from qcportal.singlepoint.test_dataset_models import test_entries, test_specs
from qcportal.tasks import TaskClaimBody, TaskReturnBody
from qcportal.utils import now_at_utc

test_mol = Molecule(symbols=["h"], geometry=[0, 0, 0])
test_uinfo = UserInfo(username="test_new_user", role="read", enabled=True)

test_project_inp_1 = SinglepointInput(
    molecule={"symbols": ["h", "h"], "geometry": [0, 0, 0, 0, 0, 2]},
    specification={
        "program": "psi4",
        "driver": "energy",
        "method": "hf",
        "basis": "sto-3g",
    },
)


@pytest.fixture(scope="function")
def roletest_snowflake(secure_snowflake_allow_read):
    # Add a test molecule, record, and dataset
    client = secure_snowflake_allow_read.client("admin_user", password=test_users["admin_user"]["pw"])

    client.add_molecules([test_mol])
    _, rids = client.add_singlepoints(
        molecules=[test_mol], program="psi4", driver="energy", method="b3lyp", basis="sto-3g"
    )
    ds = client.add_dataset("singlepoint", "testds")
    ds.add_entries(test_entries)
    ds.add_specification("test_spec", test_specs[0])

    pr = client.add_project("test_project")
    pr.link_record(rids[0], "test record", "test description", ["tag1", "tag2"])
    pr.link_dataset(ds.id)

    return secure_snowflake_allow_read


# Make functions that only take a portal client as the only argument, then
# Assemble those into a testing map
def submit_dataset(client: PortalClient):
    client.get_dataset("singlepoint", "testds").submit()


def submit_dataset_background(client: PortalClient):
    client.get_dataset("singlepoint", "testds").background_submit()


def delete_dataset_entry(client: PortalClient):
    client.get_dataset("singlepoint", "testds").delete_entries(["testent"])


def delete_dataset_specification(client: PortalClient):
    client.get_dataset("singlepoint", "testds").delete_specification("testspec")


def remove_dataset_records(client: PortalClient):
    client.get_dataset("singlepoint", "testds").remove_records(["testent"], ["testspec"])


def read_dataset(client: PortalClient):
    ds = client.get_dataset("singlepoint", "testds")
    for e, s, r in ds.iterate_records():
        ds.get_entry(e)


def get_dataset_internal_jobs(client: PortalClient):
    client.get_dataset("singlepoint", "testds").list_internal_jobs()


def get_dataset_attachments(client: PortalClient):
    _ = client.get_dataset("singlepoint", "testds").attachments


def modify_dataset_metadata(client: PortalClient):
    client.get_dataset("singlepoint", "testds").set_default_compute_tag("abc")


def create_dataset_view(client: PortalClient):
    client.get_dataset("singlepoint", "testds").create_view("a test view", {})


def activate_manager(client: PortalClient):
    # Go through the route directly
    manager_info = ManagerActivationBody(
        name_data=mname1,
        manager_version="1.23",
        username=client.username,
        programs={"prog": ["1.23"]},
        compute_tags=["*"],
    )

    return client.make_request(
        "post",
        "compute/v1/managers",
        None,
        body=manager_info,
    )


def claim_tasks(client: PortalClient):
    # Go through the route directly
    # Doesn't matter if the manager has been activated - we are looking for "Forbidden"
    body = TaskClaimBody(name_data=mname1, programs={"prog": ["1.23"]}, compute_tags=["*"], limit=1)

    client.make_request(
        "post",
        "compute/v1/tasks/claim",
        Any,
        body=body,
    )


def return_tasks(client: PortalClient):
    # Go through the route directly
    # Doesn't matter if the manager has been activated - we are looking for "Forbidden"

    body = TaskReturnBody(name_data=mname1, results_compressed={})
    client.make_request(
        "post",
        "compute/v1/tasks/return",
        Any,
        body=body,
    )


def manager_modify_heartbeat(client: PortalClient):
    # Go through the route directly
    # Doesn't matter if the manager has been activated - we are looking for "Forbidden"
    body = ManagerUpdateBody(status="active", active_tasks=10, active_cores=20, active_memory=100, total_cpu_hours=1000)

    client.make_request(
        "patch",
        f"compute/v1/managers/{mname1.fullname}",
        Any,
        body=body,
    )


def manager_modify_deactivate(client: PortalClient):
    # Go through the route directly
    # Doesn't matter if the manager has been activated - we are looking for "Forbidden"
    body = ManagerUpdateBody(
        status="inactive", active_tasks=10, active_cores=20, active_memory=100, total_cpu_hours=1000
    )

    client.make_request(
        "patch",
        f"compute/v1/managers/{mname1.fullname}",
        Any,
        body=body,
    )


def access_my_user_me(client: PortalClient):
    # Access own user, via /me
    return client.make_request("get", f"api/v1/me", UserInfo)


def access_my_user(client: PortalClient):
    # Access own user, via /users
    return client.make_request("get", f"api/v1/users/{client.username}", UserInfo)


def get_my_preferences_me(client: PortalClient):
    return client.make_request("get", f"api/v1/me/preferences", Any)


def get_my_preferences(client: PortalClient):
    return client.make_request("get", f"api/v1/users/{client.username}/preferences", Any)


def get_other_preferences(client: PortalClient):
    return client.make_request("get", f"api/v1/users/submit_user_2/preferences", Any)


def get_my_sessions_me(client: PortalClient):
    return client.make_request("get", f"api/v1/me/sessions", Any)


def get_my_sessions(client: PortalClient):
    return client.make_request("get", f"api/v1/users/{client.username}/sessions", Any)


def list_all_sessions(client: PortalClient):
    return client.make_request("get", f"api/v1/sessions", Any)


def get_other_sessions(client: PortalClient):
    return client.make_request("get", f"api/v1/users/submit_user_2", Any)


def project_list_records(client: PortalClient):
    client.get_project("test_project").record_metadata()


def project_list_datasets(client: PortalClient):
    client.get_project("test_project").dataset_metadata()


def project_link_record(client: PortalClient):
    client.get_project("test_project").link_record(9999, "name", "Desc", ["tag"])


def project_link_dataset(client: PortalClient):
    client.get_project("test_project").link_dataset(9999)


def project_add_dataset(client: PortalClient):
    client.get_project("test_project").add_dataset("singlepoint", "newdataset")


def project_add_record(client: PortalClient):
    client.get_project("test_project").add_record("test new rec", test_project_inp_1)


# fmt: off
test_function_map = {
    "get_information": PortalClient.get_server_information,
    "set_motd": partial(PortalClient.set_motd, new_motd="hi"),

    "get_access_log": [PortalClient.query_access_log,
                       PortalClient.query_access_summary],
    "delete_access_log": partial(PortalClient.delete_access_log, before=now_at_utc()),

    "get_error_log": PortalClient.query_error_log,
    "delete_error_log": partial(PortalClient.delete_error_log, before=now_at_utc()),

    "get_internal_job": [partial(PortalClient.get_internal_job, job_id=1),
                         PortalClient.query_internal_jobs],
    "modify_internal_job": partial(PortalClient.cancel_internal_job, job_id=1),
    "delete_internal_job": partial(PortalClient.delete_internal_job, job_id=1),

    "list_users": PortalClient.list_users,
    "add_user": partial(PortalClient.add_user, user_info=test_uinfo),
    "delete_user": partial(PortalClient.delete_user, username_or_id=1),

    "get_me": [access_my_user,
               access_my_user_me],
    "get_other_user": partial(PortalClient.get_user, username_or_id="submit_user_2"),

    "get_my_preferences": [get_my_preferences_me,
                           get_my_preferences],
    "get_other_preferences": get_other_preferences,

    "get_my_sessions": [get_my_sessions_me,
                        get_my_sessions],
    "list_all_sessions": list_all_sessions,
    "get_other_sessions": get_other_sessions,

    "list_groups": PortalClient.list_groups,
    "add_group": partial(PortalClient.add_group, group_info=GroupInfo(groupname="testgroup")),
    "get_group": partial(PortalClient.get_group, groupname_or_id="group1"),
    "delete_group": partial(PortalClient.delete_group, groupname_or_id="group1"),

    "add_molecule": partial(PortalClient.add_molecules, molecules=[test_mol]),
    "get_molecule": partial(PortalClient.get_molecules, molecule_ids=[1]),
    "query_molecule": PortalClient.query_molecules,
    "modify_molecule": partial(PortalClient.modify_molecule, molecule_id=1, name="new_name"),
    "delete_molecule": partial(PortalClient.delete_molecules, molecule_ids=[1]),

    "add_record": partial(PortalClient.add_singlepoints, molecules=[test_mol],  program='psi4', driver='energy', method='b3lyp', basis='sto-3g'),
    "get_record": partial(PortalClient.get_records, record_ids=[1], missing_ok=True),
    "query_record": PortalClient.query_records,
    "modify_record": [partial(PortalClient.modify_records, record_ids=[1], new_compute_tag="new_tag"),
                      partial(PortalClient.cancel_records, record_ids=[1])],
    "delete_record": partial(PortalClient.delete_records, record_ids=[1]),

    "list_datasets": PortalClient.list_datasets,
    "add_dataset": partial(PortalClient.add_dataset, dataset_type="singlepoint", name="newdataset"),
    "get_dataset": [partial(PortalClient.get_dataset_by_id, dataset_id=1),
                    partial(PortalClient.get_dataset, dataset_type="singlepoint", dataset_name="testds")],
    "read_dataset": [read_dataset,
                     get_dataset_internal_jobs,
                     get_dataset_attachments],
    "submit_dataset": [submit_dataset,
                       submit_dataset_background],
    "modify_dataset": [delete_dataset_entry,
                       delete_dataset_specification,
                       remove_dataset_records,
                       modify_dataset_metadata],
    "create_dataset_view": create_dataset_view,
    "delete_dataset": partial(PortalClient.delete_dataset, dataset_id=1, delete_records=False),

    "get_manager": partial(PortalClient.get_managers, names=['testmanager'], missing_ok=True),
    "query_manager": PortalClient.query_managers,

    "activate_manager": activate_manager,
    "modify_manager": [manager_modify_heartbeat, manager_modify_deactivate],
    "claim_tasks": claim_tasks,
    "return_tasks": return_tasks,

    "add_project": partial(PortalClient.add_project, name="test_project"),
    "get_project": [partial(PortalClient.get_project, project_name="test_project"),
                    partial(PortalClient.get_project_by_id, project_id=1)],
    "read_project": [project_list_records, project_list_datasets],
    "modify_project": [project_add_dataset, project_link_dataset, project_add_record, project_link_record],
    "delete_project": partial(PortalClient.delete_project, project_id=1),

}

test_role_permissions_map = {
    "admin": set(test_function_map.keys()),
    "maintain": {"get_information", "set_motd",
                 "get_access_log", "delete_access_log",
                 "get_error_log", "delete_error_log",
                 "get_internal_job", "modify_internal_job", "delete_internal_job",
                 "list_groups", "add_group", "get_group", "delete_group",
                 "add_molecule", "get_molecule", "query_molecule", "modify_molecule", "delete_molecule",
                 "add_record", "get_record", "query_record", "modify_record", "delete_record",
                 "list_datasets", "add_dataset", "get_dataset", "read_dataset", "submit_dataset",
                 "modify_dataset", "delete_dataset", "create_dataset_view",
                 "add_project", "get_project", "read_project", "modify_project", "delete_project",
                 "get_manager", "query_manager",
                 "get_me", "get_my_preferences", "get_my_sessions"},
    "monitor": {"get_information",
                "get_access_log", "get_error_log", "get_internal_job",
                "get_molecule", "query_molecule", "get_record", "query_record", "list_datasets",
                "get_dataset", "read_dataset", "get_manager", "query_manager",
                "get_project", "read_project",
                "get_me", "get_my_preferences", "get_my_sessions"},
    "submit": {"get_information",
               "add_molecule", "get_molecule", "query_molecule", "modify_molecule", "delete_molecule",
               "add_record", "get_record", "query_record", "modify_record", "delete_record",
               "list_datasets", "add_dataset", "get_dataset", "read_dataset", "submit_dataset",
               "modify_dataset", "delete_dataset", "create_dataset_view",
               "add_project", "get_project", "read_project", "modify_project", "delete_project",
               "get_manager", "query_manager",
               "get_me", "get_my_preferences", "get_my_sessions"},
    "read": {"get_information", "get_molecule", "query_molecule", "get_record", "query_record", "list_datasets",
             "get_dataset", "read_dataset", "get_project", "read_project", "get_manager", "query_manager",
             "get_me", "get_my_preferences", "get_my_sessions"},
    "anonymous": {"get_information", "get_molecule", "query_molecule", "get_record", "query_record", "list_datasets",
                  "get_dataset", "read_dataset", "get_project", "read_project", "get_manager", "query_manager"},
    "compute": {"get_information", "activate_manager", "modify_manager", "claim_tasks", "return_tasks",
                "get_me", "get_my_preferences", "get_my_sessions"},
}
# fmt: off

@pytest.mark.parametrize(
    "username,role",
    [
        ('admin_user', 'admin'),
        ('maintain_user', 'maintain'),
        ('monitor_user', 'monitor'),
        ('submit_user', 'submit'),
        ('read_user', 'read'),
        (None, 'anonymous'),
        ('compute_user', 'compute')
    ]
)
def test_auth_global_role_permissions(roletest_snowflake, username, role):
    if username is None:
        client = roletest_snowflake.client()
    else:
        client = roletest_snowflake.client(username, password=test_users[username]["pw"])

    role_perms = test_role_permissions_map[role]

    # Make sure our tests are sane - everything in the permissions map is something we test
    assert set(role_perms) <= set(test_function_map.keys())

    for action, f in test_function_map.items():
        allowed = action in role_perms
        print(role, action, allowed)

        if allowed:
            # Wrap in try/except - exceptions are ok, as long as they aren't "Forbidden" errors
            if isinstance(f, list):
                for x in f:
                    try:
                        x(client)
                    except Exception as err:
                        assert "Forbidden" not in str(err)
            else:
                try:
                    f(client)
                except Exception as err:
                    assert "Forbidden" not in str(err)
        else:
            if isinstance(f, list):
                for x in f:
                    with pytest.raises(PortalRequestError, match="Forbidden"):
                        x(client)
            else:
                with pytest.raises(PortalRequestError, match="Forbidden"):
                    f(client)


def test_auth_global_no_unauth_read(secure_snowflake):
    with pytest.raises(PortalRequestError, match="Server requires login"):
        secure_snowflake.client()

    # Fake a client - constructing normally results in trying to get information
    client = secure_snowflake.client("admin_user", password=test_users["admin_user"]["pw"])
    client.username = None
    client._password = None
    client._jwt_access_exp = None
    client._jwt_refresh_exp = None
    client._jwt_refresh_token = None
    client._jwt_access_token = None
    client._req_session.headers.pop("Authorization")

    for action, f in test_function_map.items():
        if isinstance(f, list):
            for x in f:
                with pytest.raises(PortalRequestError, match="Server requires login"):
                    x(client)
        else:
            with pytest.raises(PortalRequestError, match="Server requires login"):
                f(client)

@pytest.mark.parametrize(
    "username,role",
    [
        ('admin_user', 'admin'),
        ('maintain_user', 'maintain'),
        ('monitor_user', 'monitor'),
        ('submit_user', 'submit'),
        ('read_user', 'read'),
        (None, 'anonymous'),
        ('compute_user', 'compute')
    ]
)
def test_auth_global_user_management(secure_snowflake_allow_read, username, role):
    # We need userinfo for a completely different user
    client0 = secure_snowflake_allow_read.client("submit_user_2", password=test_users["submit_user_2"]["pw"])
    other_uinfo = client0.get_user()

    if username is None:
        client = secure_snowflake_allow_read.client()
    else:
        client = secure_snowflake_allow_read.client(username, password=test_users[username]["pw"])

    if username is None:
        # cannot access /me
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("get", f"api/v1/me", UserInfo)
    else:
        client.make_request("get", f"api/v1/me", UserInfo)

    # All can change own password, except if no user
    if username is None:
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/me/password", Any, body="new_password")
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/me/password", Any, body_model=Optional[str], body=None)
    else:
        client.make_request("put", f"api/v1/me/password", Any, body="new_password")
        client.make_request("put", f"api/v1/me/password", Any, body_model=Optional[str], body=None)
        client.make_request("put", f"api/v1/users/{username}/password", Any, body="new_password")
        client.make_request("put", f"api/v1/users/{username}/password", Any, body_model=Optional[str], body=None)

    # All can access or change own preferences, except if no user
    if username is None:
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("get", f"api/v1/me/preferences", dict)
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/me/preferences", None, body_model=dict, body={})
    else:
        client.make_request("get", f"api/v1/me/preferences", dict)
        client.make_request("put", f"api/v1/me/preferences", None, body_model=dict, body={})
        client.make_request("put", f"api/v1/users/{username}/preferences", None, body_model=dict, body={})

    if role != 'admin':
        # cannot access or modify other users
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("get", f"api/v1/users/submit_user_2", UserInfo)
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("patch", f"api/v1/users", UserInfo, body=other_uinfo)
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/users/submit_user_2/password", Any, body="new_password")
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/users/submit_user_2/password", str, body_model=Optional[str], body=None)
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("get", f"api/v1/users/submit_user_2/preferences", dict)
        with pytest.raises(PortalRequestError, match="Forbidden"):
            client.make_request("put", f"api/v1/users/submit_user_2/preferences", None, body_model=dict, body={})
    else:
        client.make_request("get", f"api/v1/users/submit_user_2", UserInfo)
        client.make_request("patch", f"api/v1/users", UserInfo, body=other_uinfo)
        client.make_request("put", f"api/v1/users/submit_user_2/password", Any, body="new_password")
        client.make_request("put", f"api/v1/users/submit_user_2/password", str, body_model=Optional[str], body=None)
        client.make_request("get", f"api/v1/users/submit_user_2/preferences", dict)
        client.make_request("put", f"api/v1/users/submit_user_2/preferences", None, body_model=dict, body={})

@pytest.mark.parametrize("as_admin", [True, False])
def test_auth_protected_endpoints(snowflake, as_admin):
    # Cannot add or modify users/groups when security is disabled
    uinfo = UserInfo(username='test', role="read", enabled=True)
    ginfo = GroupInfo(groupname="testg")

    snowflake.create_users()

    if as_admin:
        client: PortalClient = snowflake.client(username="admin_user", password=test_users["admin_user"]["pw"])
    else:
        client: PortalClient = snowflake.client()

    # Can't add user
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.add_user(uinfo)

    # Can't get own user or another user
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("get", f"api/v1/me", UserInfo)

    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("get", f"api/v1/users/submit_user", UserInfo)

    # Can't modify a user
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("patch", f"api/v1/users", UserInfo, body=uinfo)

    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("patch", f"api/v1/me", UserInfo, body=uinfo)

    # Can't change passwords
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("put", f"api/v1/me/password", Any, body="new_password")
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.make_request("put", f"api/v1/users/submit_user/password", Any, body="new_password")

    # Can't delete users
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.delete_user("submit_user")

    # Can't add or delete groups
    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.add_group(ginfo)

    with pytest.raises(PortalRequestError, match="Cannot access.*security disabled"):
        client.delete_group("group1")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcportal.auth.models import GroupInfo
from qcportal.exceptions import (
    UserManagementError,
    InvalidGroupnameError,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket

invalid_groupnames = ["\x00", "ab\x00cd", "a user", ""]


def test_group_socket_add_get(storage_socket: SQLAlchemySocket):
    ginfo = GroupInfo(
        groupname="group1",
        description="Test group 1",
    )
    storage_socket.groups.add(ginfo)

    # Do we get the same data back?
    # The initial userinfo doesn't contain the id
    ginfo2 = storage_socket.groups.get("group1")
    ginfo2.pop("id")
    assert GroupInfo(**ginfo2) == ginfo

    # Get by id
    ginfo2 = storage_socket.groups.get("group1")
    ginfo3 = storage_socket.groups.get(ginfo2["id"])
    assert ginfo2 == ginfo3


def test_group_socket_add_duplicate(storage_socket: SQLAlchemySocket):
    ginfo = GroupInfo(
        groupname="group1",
        description="Test group 1",
    )
    storage_socket.groups.add(ginfo)

    # Duplicate should result in an exception
    ginfo2 = GroupInfo(
        groupname="group1",
        description="Another Test group 1",
    )
    with pytest.raises(UserManagementError, match=r"Group.*already exists"):
        storage_socket.groups.add(ginfo2)


def test_group_socket_add_with_id(storage_socket: SQLAlchemySocket):
    # Should not be able to add a group with the id set
    ginfo = GroupInfo(
        id=1234,
        groupname="group1",
        description="Test group 1",
    )

    with pytest.raises(UserManagementError, match=r"id was given as part"):
        storage_socket.groups.add(ginfo)


def test_group_socket_list(storage_socket: SQLAlchemySocket):
    all_groups = []
    for i in range(20):
        ginfo = GroupInfo(groupname=f"group_{i}", description=f"Test group {i}")
        storage_socket.groups.add(ginfo)
        all_groups.append(ginfo)

    group_lst = storage_socket.groups.list()
    group_lst_model = [GroupInfo(**x) for x in group_lst]

    # Sort both lists by username
    all_groups = sorted(all_groups, key=lambda x: x.groupname)
    group_lst_model = sorted(group_lst_model, key=lambda x: x.groupname)

    d1 = [x.dict(exclude={"id"}) for x in all_groups]
    d2 = [x.dict(exclude={"id"}) for x in group_lst_model]
    assert d1 == d2


def test_group_socket_delete(storage_socket: SQLAlchemySocket):
    ginfo1 = GroupInfo(
        groupname="group1",
        description="Test group 1",
    )
    ginfo2 = GroupInfo(
        groupname="group2",
        description="Test group 1",
    )
    storage_socket.groups.add(ginfo1)
    storage_socket.groups.add(ginfo2)

    gid1 = storage_socket.groups.get("group1")["id"]
    gid2 = storage_socket.groups.get("group2")["id"]

    # Raises exception on error
    storage_socket.groups.delete("group1")
    storage_socket.groups.delete(gid2)

    with pytest.raises(UserManagementError, match=r"Group.*does not exist"):
        storage_socket.groups.get("group1")
    with pytest.raises(UserManagementError, match=r"Group.*does not exist"):
        storage_socket.groups.get(gid1)
    with pytest.raises(UserManagementError, match=r"Group.*does not exist"):
        storage_socket.groups.get("group2")
    with pytest.raises(UserManagementError, match=r"Group.*does not exist"):
        storage_socket.groups.get(gid2)


def test_group_socket_use_invalid_groupname(storage_socket: SQLAlchemySocket):
    # TESTING INVALID GROUP NAMES #

    for groupname in invalid_groupnames:
        # Normally, GroupInfo prevents bad usernames. But the socket also checks, as a last resort
        # So we have to bypass the GroupInfo check with construct()
        ginfo = GroupInfo.construct(
            groupname=groupname,
            description="Test group 1",
        )

        with pytest.raises(InvalidGroupnameError):
            storage_socket.groups.add(ginfo)

        with pytest.raises(InvalidGroupnameError):
            storage_socket.groups.get(groupname)

        with pytest.raises(InvalidGroupnameError):
            storage_socket.groups.delete(groupname)

    # numeric groupnames are not allowed for some actions
    ginfo2 = GroupInfo.construct(
        groupname="123456789",
        description="Test group 1",
    )

    with pytest.raises(InvalidGroupnameError):
        storage_socket.groups.add(ginfo2)



from flask import current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.compute_v1.blueprint import compute_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import LimitExceededError
from qcportal.tasks import TaskClaimBody, TaskReturnBody
from qcportal.utils import calculate_limit


# WRITE action is not a mistake. Claim does some changes to the DB, so require
# a bit more than read
@compute_v1.route("/tasks/claim", methods=["POST"])
@wrap_global_route("tasks", "claim")
def claim_tasks_v1(body_data: TaskClaimBody):
    """Claims tasks from the task queue"""

    # check here, but also in the socket
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.manager_tasks_claim

    return storage_socket.tasks.claim_tasks(
        manager_name=body_data.name_data.fullname,
        compute_tags=body_data.compute_tags,
        programs=body_data.programs,
        limit=calculate_limit(max_limit, body_data.limit),
    )


@compute_v1.route("/tasks/return", methods=["POST"])
@wrap_global_route("tasks", "return")
def return_tasks_v1(body_data: TaskReturnBody):
    """Return finished tasks"""

    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.manager_tasks_claim
    if len(body_data.results_compressed) > max_limit:
        raise LimitExceededError(f"Attempted to return too many results - limit is {max_limit}")

    return storage_socket.tasks.update_finished(
        manager_name=body_data.name_data.fullname, results_compressed=body_data.results_compressed
    )



import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.schema import DDLElement


# Taken from https://github.com/sqlalchemy/sqlalchemy/wiki/Views


class CreateView(DDLElement):
    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(DDLElement):
    def __init__(self, name):
        self.name = name


@compiler.compiles(CreateView)
def _create_view(element, compiler, **kw):
    return "CREATE VIEW %s AS %s" % (
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


@compiler.compiles(DropView)
def _drop_view(element, compiler, **kw):
    return "DROP VIEW %s" % (element.name)


def view_exists(ddl, target, connection, **kw):
    return ddl.name in sa.inspect(connection).get_view_names()


def view_doesnt_exist(ddl, target, connection, **kw):
    return not view_exists(ddl, target, connection, **kw)


def view(name, metadata, selectable):

    t = sa.table(
        name,
        *(sa.Column(c.name, c.type, primary_key=c.primary_key) for c in selectable.selected_columns),
    )
    t.primary_key.update(c for c in t.c if c.primary_key)

    sa.event.listen(
        metadata,
        "after_create",
        CreateView(name, selectable).execute_if(callable_=view_doesnt_exist),
    )
    sa.event.listen(
        metadata,
        "before_drop",
        DropView(name).execute_if(callable_=view_exists),
    )
    return t



from __future__ import annotations

import logging
import traceback
from collections import defaultdict
from typing import TYPE_CHECKING

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcelemental.models import FailedOperation
from sqlalchemy import select, update
from sqlalchemy.dialects.postgresql import array
from sqlalchemy.orm import selectinload, load_only, lazyload, undefer

from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.record_db_models import BaseRecordORM, RecordComputeHistoryORM, OutputStoreORM
from qcportal.all_results import AllResultTypes
from qcportal.compression import CompressionEnum, compress
from qcportal.compression import decompress
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerStatusEnum
from qcportal.metadata_models import TaskReturnMetadata
from qcportal.record_models import RecordStatusEnum
from qcportal.utils import calculate_limit, now_at_utc
from .db_models import TaskQueueORM
from .reset_logic import should_reset

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Any


class TaskSocket:
    """
    Socket for managing tasks
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self._tasks_claim_limit = root_socket.qcf_config.api_limits.manager_tasks_claim
        self._strict_compute_tags = root_socket.qcf_config.strict_compute_tags

    def update_finished(
        self, manager_name: str, results_compressed: Dict[int, bytes], *, session: Optional[Session] = None
    ) -> TaskReturnMetadata:
        """
        Insert data from finished calculations into the database

        Parameters
        ----------
        manager_name
            The name of the manager submitting the results
        results_compressed
            Results (in QCSchema format), with the task_id as the key
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        all_task_ids = list(results_compressed.keys())

        self._logger.info("Received completed tasks from {}.".format(manager_name))
        self._logger.info("    Task ids: " + " ".join(str(x) for x in all_task_ids))

        tasks_success: List[int] = []
        tasks_failures: List[int] = []
        tasks_rejected: List[Tuple[int, str]] = []

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == manager_name)
            manager: Optional[ComputeManagerORM] = session.execute(stmt).scalar_one_or_none()

            if manager is None:
                self._logger.warning(f"Manager {manager_name} does not exist, but is trying to return tasks. Ignoring.")
                raise ComputeManagerError(f"Manager {manager_name} does not exist")

            if manager.status != ManagerStatusEnum.active:
                self._logger.warning(f"Manager {manager_name} is not active. Ignoring...")
                raise ComputeManagerError(f"Manager {manager_name} is not active")

            # For automatic resetting
            to_be_reset: List[int] = []

            # We load basic record & task info for all returned tasks with row-level locking.
            # This lock is released on commit or rollback.
            all_task_ids = list(results_compressed.keys())
            stmt = select(
                TaskQueueORM.id,
                BaseRecordORM.id,
                BaseRecordORM.record_type,
                BaseRecordORM.status,
                BaseRecordORM.manager_name,
            )
            stmt = stmt.join(TaskQueueORM, TaskQueueORM.record_id == BaseRecordORM.id)
            stmt = stmt.where(TaskQueueORM.id.in_(all_task_ids))
            stmt = stmt.with_for_update(skip_locked=False)

            all_record_info = session.execute(stmt).all()
            all_record_info = {x[0]: x[1:] for x in all_record_info}

            for task_id, result_compressed in results_compressed.items():

                record_info = all_record_info.get(task_id, None)

                #################################################################
                # Perform some checks for consistency
                # These are simple rejections and don't modify anything
                # record-related in the database
                #################################################################
                if record_info is None:
                    self._logger.warning(f"Task id {task_id} does not exist in the task queue")
                    tasks_rejected.append((task_id, "Task does not exist in the task queue"))
                    continue

                record_id, record_type, record_status, record_manager_name = record_info

                # Is the task in the running state
                # If so, do not attempt to modify the task queue. Just move on
                if record_status != RecordStatusEnum.running:
                    self._logger.warning(f"Record {record_id} (task {task_id}) is not in a running state")
                    tasks_rejected.append((task_id, "Task is not in a running state"))
                    continue

                # Was the manager that sent the data the one that was assigned?
                # If so, do not attempt to modify the task queue. Just move on
                if record_manager_name != manager_name:
                    self._logger.warning(
                        f"Record {record_id} (task {task_id}) claimed by {record_manager_name}, not {manager_name}"
                    )
                    tasks_rejected.append((task_id, "Task is claimed by another manager"))
                    continue

                result_dict = decompress(result_compressed, CompressionEnum.zstd)
                result = pydantic.parse_obj_as(AllResultTypes, result_dict)

                notify_status = None

                ##################################################################
                # The rest of the checks are done in a try/except block because
                # they are much more complicated and can result in exceptions
                # which should be handled
                ##################################################################

                try:
                    savepoint = session.begin_nested()

                    # Failed task returning FailedOperation
                    if result.success is False and isinstance(result, FailedOperation):
                        self.root_socket.records.update_failed_task(session, record_id, result, manager_name)

                        notify_status = RecordStatusEnum.error
                        tasks_failures.append(task_id)

                        # Should we automatically reset?
                        if self.root_socket.qcf_config.auto_reset.enabled:
                            # TODO - Move to update_failed_task?
                            stmt = select(BaseRecordORM)

                            # We will need to load the compute history and outputs to determine if we should reset
                            stmt = stmt.options(
                                selectinload(BaseRecordORM.compute_history).options(
                                    selectinload(RecordComputeHistoryORM.outputs).options(undefer(OutputStoreORM.data))
                                )
                            )
                            stmt = stmt.where(BaseRecordORM.id == record_id)
                            record_orm = session.execute(stmt).scalar_one()
                            if should_reset(record_orm, self.root_socket.qcf_config.auto_reset):
                                to_be_reset.append(record_id)

                    elif result.success is not True:
                        # QCEngine should always return either FailedOperation, or some result with success == True
                        msg = f"Unexpected return from manager for task {task_id}/base result {record_id}: Returned success != True, but not a FailedOperation"
                        error = {"error_type": "internal_fractal_error", "error_message": msg}
                        failed_op = FailedOperation(error=error, success=False)

                        self.root_socket.records.update_failed_task(session, record_id, failed_op, manager_name)
                        notify_status = RecordStatusEnum.error

                        self._logger.error(msg)
                        tasks_rejected.append((task_id, "Returned success=False, but not a FailedOperation"))

                    # Manager returned a full, successful result
                    else:
                        self.root_socket.records.update_completed(session, record_id, record_type, result, manager_name)

                        notify_status = RecordStatusEnum.complete
                        tasks_success.append(task_id)

                    savepoint.commit()  # Release the savepoint (doesn't actually fully commit)

                except Exception:
                    # We have no idea what was added or is pending for removal
                    # So rollback the transaction to the most recent commit
                    savepoint.rollback()
                    savepoint = session.begin_nested()

                    msg = "Internal FractalServer Error:\n" + traceback.format_exc()
                    error = {"error_type": "internal_fractal_error", "error_message": msg}
                    failed_op = FailedOperation(error=error, success=False)

                    self.root_socket.records.update_failed_task(session, record_id, failed_op, manager_name)
                    notify_status = RecordStatusEnum.error

                    self._logger.error(msg)
                    tasks_rejected.append((task_id, "Internal server error"))

                    savepoint.commit()

                finally:
                    # Send notifications that tasks were completed
                    # Notifications are sent after the transaction is committed
                    if notify_status is not None:
                        self.root_socket.notify_finished_watch(record_id, notify_status)

            session.commit()

            # Update the stats for the manager
            manager.successes += len(tasks_success)
            manager.failures += len(tasks_failures)
            manager.rejected += len(tasks_rejected)

            # Mark that we have heard from the manager
            manager.modified_on = now_at_utc()

            # Automatically reset ones that should be reset
            if self.root_socket.qcf_config.auto_reset.enabled and to_be_reset:
                self._logger.info(f"Auto resetting {len(to_be_reset)} records: ")
                for record_id in to_be_reset:
                    self._logger.info(f"    {record_id}")

                self.root_socket.records.reset(to_be_reset, session=session)

        self._logger.info(
            "Processed {} returned tasks ({} successful, {} failed, {} rejected).".format(
                len(results_compressed), len(tasks_success), len(tasks_failures), len(tasks_rejected)
            )
        )

        return TaskReturnMetadata(rejected_info=tasks_rejected, accepted_ids=(tasks_success + tasks_failures))

    def claim_tasks(
        self,
        manager_name: str,
        programs: Dict[str, List[str]],
        compute_tags: List[str],
        limit: Optional[int] = None,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        """Claim/assign tasks for a manager

        Given tags and available programs/procedures on the manager, obtain
        waiting tasks to run.

        Parameters
        ----------
        manager_name
            Name of the manager requesting tasks
        compute_tags
            Subset of tags to claim tasks for
        limit
            Maximum number of tasks that the manager can claim
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # Normally, checking limits is done in the route code. However, we really do not want a manager
        # to claim absolutely everything. So double check here
        limit = calculate_limit(self._tasks_claim_limit, limit)

        # Force given tags and programs to be lower case
        compute_tags = [tag.lower() for tag in compute_tags]
        programs = {key.lower(): [x.lower() for x in value] for key, value in programs.items()}

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == manager_name)
            stmt = stmt.with_for_update(skip_locked=False)
            manager: Optional[ComputeManagerORM] = session.execute(stmt).scalar_one_or_none()

            if manager is None:
                self._logger.warning(f"Manager {manager_name} does not exist! Will not give it tasks")
                raise ComputeManagerError("Manager does not exist!")
            elif manager.status != ManagerStatusEnum.active:
                self._logger.warning(f"Manager {manager_name} exists but is not active! Will not give it tasks")
                raise ComputeManagerError("Manager is not active!")

            # Remove tags & programs that we didn't say we handled
            # (order is important for tags)
            search_programs = array(p for p in programs.keys() if p in manager.programs.keys())
            search_tags = [x for x in compute_tags if x in manager.compute_tags]

            if len(search_programs) == 0:
                self._logger.warning(f"Manager {manager_name} did not send any valid programs to claim")
                raise ComputeManagerError(f"Manager {manager_name} did not send any valid programs to claim")

            if len(search_tags) == 0:
                self._logger.warning(f"Manager {manager_name} did not send any valid queue tags to claim")
                raise ComputeManagerError(f"Manager {manager_name} did not send any valid queue tags to claim")

            found: Dict[int, Dict[str, Any]] = {}
            return_order: List[int] = []  # Order of task ids

            for tag in search_tags:
                new_limit = limit - len(found)

                # Have we found all we needed to find
                # (should always be >= 0, but can never be too careful. If it wasn't this could be an infinite loop)
                if new_limit <= 0:
                    break

                # Find tasks/base result:
                #   1. Status is waiting
                #   2. Whose required programs I am able to match
                #   3. Whose tags I am able to compute
                # with_for_update locks the rows. skip_locked=True makes it skip already-locked rows
                # (possibly from another process)
                # Also, load the base_result object so we can update stuff there (status)
                # TODO - we only test for the presence of the available_programs in the requirements. Eventually
                #        we want to then verify the versions

                # This is a very tricky query. We only want to load two columns of BaseRecord, but we need to join.
                # Sqlalchemy likes to load more columns if there is a relationship. So we join outside of the
                # relationship (ie, NOT TaskQueueORM.record) and then load the columns we want.
                stmt = select(TaskQueueORM, BaseRecordORM).join(
                    BaseRecordORM, BaseRecordORM.id == TaskQueueORM.record_id
                )

                stmt = stmt.filter(TaskQueueORM.available == True)
                stmt = stmt.filter(search_programs.contains(TaskQueueORM.required_programs))
                stmt = stmt.options(load_only(BaseRecordORM.id, BaseRecordORM.record_type))
                stmt = stmt.options(lazyload(BaseRecordORM.creator_user))

                # Order by priority, then date (earliest first)
                # The sort_date usually comes from the created_on of the record, or the created_on of the record's parent service
                stmt = stmt.order_by(
                    TaskQueueORM.compute_priority.desc(), TaskQueueORM.sort_date.asc(), TaskQueueORM.id.asc()
                )

                # If tag is "*" (and strict_compute_tags is False), then the manager can pull anything
                # If tag is "*" and strict_compute_tags is enabled, only pull tasks with tag == '*'
                if tag != "*" or self._strict_compute_tags:
                    stmt = stmt.filter(TaskQueueORM.compute_tag == tag)

                # Skip locked rows - They may be in the process of being claimed by someone else
                stmt = stmt.limit(new_limit).with_for_update(of=[BaseRecordORM, TaskQueueORM], skip_locked=True)

                new_items = session.execute(stmt).all()

                # Update all the task records to reflect this manager claiming them
                task_ids = [x[0].id for x in new_items]
                record_ids = [x[1].id for x in new_items]

                return_order.extend(task_ids)  # Keep the order as returned from the db
                stmt = update(TaskQueueORM).where(TaskQueueORM.id.in_(task_ids)).values(available=False)
                session.execute(stmt)

                stmt = (
                    update(BaseRecordORM)
                    .where(BaseRecordORM.id.in_(record_ids))
                    .values(status=RecordStatusEnum.running, manager_name=manager_name, modified_on=now_at_utc())
                )
                session.execute(stmt)

                # Store in dict form for returning, but no need to store the info from the base record
                # Also, retrieve the actual function kwargs. Eventually we may want the managers
                # to retrieve the kwargs themselves
                tasks_to_generate = defaultdict(list)
                task_updates = []

                # Find what tasks need their function and kwargs generated
                # Otherwise, just add them to the returned list
                for task_orm, record_orm in new_items:
                    if task_orm.function is None:
                        tasks_to_generate[record_orm.record_type].append(task_orm)
                    else:
                        found[task_orm.id] = task_orm.model_dict(exclude=["record"])

                # Create the task data on the fly if it doesn't exist
                for record_type, tasks_orm in tasks_to_generate.items():
                    record_socket = self.root_socket.records.get_socket(record_type)

                    record_ids = [task_orm.record_id for task_orm in tasks_orm]
                    task_specs = record_socket.generate_task_specifications(session, record_ids)

                    for task_orm, task_spec in zip(tasks_orm, task_specs):
                        task_dict = task_orm.model_dict(exclude=["record"])

                        kwargs = task_spec["function_kwargs"]
                        kwargs_compressed, _, _ = compress(kwargs, CompressionEnum.zstd)

                        # Add this to the orm for any future managers claiming this task
                        task_updates.append(
                            {
                                "id": task_orm.id,
                                "function": task_spec["function"],
                                "function_kwargs_compressed": kwargs_compressed,
                            }
                        )

                        # But just use what we created when returning to this manager
                        task_dict["function"] = task_spec["function"]
                        task_dict["function_kwargs_compressed"] = kwargs_compressed
                        found[task_orm.id] = task_dict

                # Update the task records with the function and kwargs
                if task_updates:
                    session.execute(update(TaskQueueORM), task_updates)

                session.flush()

            manager.claimed += len(found)

            # Mark that we have heard from the manager
            manager.modified_on = now_at_utc()

            self._logger.info(f"Manager {manager_name} has claimed {len(found)} new tasks")

        return [found[i] for i in return_order]



"""Remove compute manager log

Revision ID: 75b80763e901
Revises: f31c7897345f
Create Date: 2024-05-06 10:08:30.711531

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "75b80763e901"
down_revision = "f31c7897345f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_compute_manager_log_manager_id", table_name="compute_manager_log")
    op.drop_table("compute_manager_log")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "compute_manager_log",
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column("manager_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("timestamp", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=False),
        sa.Column("successes", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("claimed", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("failures", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("total_cpu_hours", sa.DOUBLE_PRECISION(precision=53), autoincrement=False, nullable=False),
        sa.Column("active_tasks", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("active_cores", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("active_memory", sa.DOUBLE_PRECISION(precision=53), autoincrement=False, nullable=False),
        sa.Column("rejected", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(
            ["manager_id"], ["compute_manager.id"], name="compute_manager_log_manager_id_fkey", ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", name="compute_manager_log_pkey"),
    )
    op.create_index("ix_compute_manager_log_manager_id", "compute_manager_log", ["manager_id"], unique=False)
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    String,
    UniqueConstraint,
    CheckConstraint,
    Index,
    Boolean,
    event,
    DDL,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.optimization.record_db_models import (
    OptimizationRecordORM,
    OptimizationSpecificationORM,
    OptimizationTrajectoryORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM, SinglepointRecordORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class NEBOptimizationsORM(BaseORM):
    __tablename__ = "neb_optimizations"

    neb_id = Column(Integer, ForeignKey("neb_record.id", ondelete="cascade"), primary_key=True)
    optimization_id = Column(Integer, ForeignKey(OptimizationRecordORM.id), primary_key=True)
    position = Column(Integer, primary_key=True)
    ts = Column(Boolean, primary_key=True)
    optimization_record = relationship(OptimizationRecordORM)

    __table_args__ = (Index("ix_neb_optimizations_optimization_id", "optimization_id"),)

    _qcportal_model_excludes = ["neb_id"]


class NEBSinglepointsORM(BaseORM):
    __tablename__ = "neb_singlepoints"

    neb_id = Column(Integer, ForeignKey("neb_record.id", ondelete="cascade"), primary_key=True)
    singlepoint_id = Column(Integer, ForeignKey(SinglepointRecordORM.id), primary_key=True)
    chain_iteration = Column(Integer, primary_key=True)
    position = Column(Integer, primary_key=True)
    singlepoint_record = relationship(SinglepointRecordORM)

    __table_args__ = (Index("ix_neb_singlepoints_singlepoint_id", "singlepoint_id"),)

    _qcportal_model_excludes = ["neb_id"]


class NEBInitialchainORM(BaseORM):
    __tablename__ = "neb_initialchain"

    neb_id = Column(Integer, ForeignKey("neb_record.id", ondelete="cascade"), primary_key=True)
    molecule_id = Column(Integer, ForeignKey(MoleculeORM.id, ondelete="cascade"), primary_key=True)
    position = Column(Integer, primary_key=True)

    molecule = relationship(MoleculeORM)

    _qcportal_model_excludes = ["neb_id"]


class NEBSpecificationORM(BaseORM):
    """
    Table for storing NEB specifications
    """

    __tablename__ = "neb_specification"

    id = Column(Integer, primary_key=True)

    program = Column(String(100), nullable=False)
    specification_hash = Column(String, nullable=False)

    singlepoint_specification_id = Column(Integer, ForeignKey(QCSpecificationORM.id), nullable=False)
    singlepoint_specification = relationship(QCSpecificationORM, lazy="joined", uselist=False)

    optimization_specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=True)
    optimization_specification = relationship(OptimizationSpecificationORM, lazy="joined")

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "specification_hash",
            "singlepoint_specification_id",
            "optimization_specification_id",
            name="ux_neb_specification_keys",
            postgresql_nulls_not_distinct=True,
        ),
        Index("ix_neb_specification_program", "program"),
        Index("ix_neb_specification_singlepoint_specification_id", "singlepoint_specification_id"),
        Index("ix_neb_specification_optimization_specification_id", "optimization_specification_id"),
        CheckConstraint("program = LOWER(program)", name="ck_neb_specification_program_lower"),
    )

    # TODO - protocols will eventually be in the model
    _qcportal_model_excludes = [
        "id",
        "specification_hash",
        "singlepoint_specification_id",
        "optimization_specification_id",
        "protocols",
    ]

    @property
    def short_description(self) -> str:
        return f"{self.program}~{self.singlepoint_specification.short_description}"


class NEBRecordORM(BaseRecordORM):
    __tablename__ = "neb_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(NEBSpecificationORM.id), nullable=False)
    specification = relationship(NEBSpecificationORM, lazy="selectin")

    initial_chain = relationship(
        NEBInitialchainORM,
        order_by=NEBInitialchainORM.position,
        collection_class=ordering_list("position"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    singlepoints = relationship(
        NEBSinglepointsORM,
        order_by=[NEBSinglepointsORM.chain_iteration, NEBSinglepointsORM.position],
        collection_class=ordering_list("position"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    optimizations = relationship(
        NEBOptimizationsORM,
        order_by=NEBOptimizationsORM.position,
        collection_class=ordering_list("position"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "neb",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseRecordORM.model_dict(self, exclude)

        # Return initial molecule or just the ids, depending on what we have
        if "initial_chain" in d:
            init_chain = d.pop("initial_chain")
            d["initial_chain_molecule_ids"] = [x["molecule_id"] for x in init_chain]
            if "molecule" in init_chain[0]:
                d["initial_chain"] = [x["molecule"] for x in init_chain]

        if "optimizations" in d:
            optimizations = d.pop("optimizations")

            opt_dict = {}
            for opt in optimizations:
                if opt["ts"]:
                    opt_dict["transition"] = opt
                elif opt["position"] == 0:
                    opt_dict["initial"] = opt
                else:
                    opt_dict["final"] = opt

            d["optimizations"] = opt_dict

        return d

    @property
    def short_description(self) -> str:
        n_mols = len(self.initial_chain)
        return f'{n_mols}*{self.initial_chain[0].molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL("""
    CREATE TRIGGER qca_neb_record_delete_base_tr
    AFTER DELETE ON neb_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """)

event.listen(NEBRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql")))

record_direct_children_select = [
    select(BaseRecordORM.id.label("parent_id"), NEBSinglepointsORM.singlepoint_id.label("child_id"))
    .join(NEBSinglepointsORM, BaseRecordORM.id == NEBSinglepointsORM.neb_id)
    .where(BaseRecordORM.record_type == "neb"),
    select(BaseRecordORM.id.label("parent_id"), NEBOptimizationsORM.optimization_id.label("child_id"))
    .join(NEBOptimizationsORM, BaseRecordORM.id == NEBOptimizationsORM.neb_id)
    .where(BaseRecordORM.record_type == "neb"),
]

record_children_select = [
    *record_direct_children_select,
    select(BaseRecordORM.id.label("parent_id"), OptimizationTrajectoryORM.singlepoint_id.label("child_id"))
    .join(NEBOptimizationsORM, BaseRecordORM.id == NEBOptimizationsORM.neb_id)
    .join(OptimizationTrajectoryORM, OptimizationTrajectoryORM.optimization_id == NEBOptimizationsORM.optimization_id)
    .where(BaseRecordORM.record_type == "neb"),
]



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, TIMESTAMP, String, Index, Enum, UniqueConstraint, ForeignKey
from sqlalchemy import DDL, event
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.orm import relationship

from qcfractal.components.auth.db_models import UserIDMapSubquery, UserORM
from qcfractal.db_socket import BaseORM
from qcportal.internal_jobs.models import InternalJobStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from typing import Optional, Iterable, Dict, Any


class InternalJobORM(BaseORM):
    __tablename__ = "internal_jobs"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    status = Column(Enum(InternalJobStatusEnum), nullable=False)
    added_date = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    scheduled_date = Column(TIMESTAMP(timezone=True), nullable=False)
    started_date = Column(TIMESTAMP(timezone=True))
    last_updated = Column(TIMESTAMP(timezone=True))
    ended_date = Column(TIMESTAMP(timezone=True))
    runner_hostname = Column(String)
    runner_uuid = Column(String)

    progress = Column(Integer, nullable=False, default=0)
    progress_description = Column(String, nullable=True)

    function = Column(String, nullable=False)
    kwargs = Column(JSON, nullable=False)

    after_function = Column(String, nullable=True)
    after_function_kwargs = Column(JSON(none_as_null=True), nullable=True)

    repeat_delay = Column(Integer, nullable=True)

    result = Column(JSON)
    user_id = Column(Integer, ForeignKey(UserORM.id, ondelete="cascade"), nullable=True)

    user = relationship(
        UserIDMapSubquery,
        foreign_keys=[user_id],
        primaryjoin="InternalJobORM.user_id == UserIDMapSubquery.id",
        lazy="selectin",
        viewonly=True,
    )

    # Nullable column with unique constraint. If a unique_name is specified,
    # it must be unique. null != null always
    unique_name = Column(String, nullable=True)

    # If this job is part of a serial group (only one may run at a time)
    serial_group = Column(String, nullable=True)

    __table_args__ = (
        Index("ix_internal_jobs_added_date", "added_date", postgresql_using="brin"),
        Index("ix_internal_jobs_scheduled_date", "scheduled_date", postgresql_using="brin"),
        Index("ix_internal_jobs_last_updated", "last_updated", postgresql_using="brin"),
        Index("ix_internal_jobs_status", "status"),
        Index("ix_internal_jobs_name", "name"),
        Index("ix_internal_jobs_user_id", "user_id"),
        UniqueConstraint("unique_name", name="ux_internal_jobs_unique_name"),
        # Enforces only one running per serial group
        Index(
            "ux_internal_jobs_status_serial_group",
            "status",
            "serial_group",
            unique=True,
            postgresql_where=(status == InternalJobStatusEnum.running),
        ),
    )

    _qcportal_model_excludes = ["unique_name", "user_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)
        d["user"] = self.user.username if self.user is not None else None
        return d


# Function that sends a postgres NOTIFY to internal job workers
# (always do notify, even if the job is in the future. The worker can calculate
# the time difference and sleep until the job is ready)
_insert_internal_job_triggerfunc = DDL(
    """
    CREATE OR REPLACE FUNCTION public.qca_internal_jobs_notify()
    RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
        BEGIN
          PERFORM pg_notify('check_internal_jobs', '');
          RETURN NEW;
        END
        $_$
    ;
"""
)

# Trigger the above function whenever a new internal job is added
_insert_internal_job_trigger = DDL(
    """
    CREATE TRIGGER qca_internal_jobs_insert_tr
    AFTER INSERT ON internal_jobs
    FOR EACH ROW EXECUTE PROCEDURE qca_internal_jobs_notify();
    """
)

event.listen(
    InternalJobORM.__table__, "after_create", _insert_internal_job_triggerfunc.execute_if(dialect=("postgresql"))
)
event.listen(InternalJobORM.__table__, "after_create", _insert_internal_job_trigger.execute_if(dialect=("postgresql")))



from __future__ import annotations

from typing import TYPE_CHECKING

from qcelemental.models import FailedOperation

from qcfractal.components.torsiondrive.testing_helpers import (
    submit_procedure_data as submit_td_procedure_data,
    generate_task_key as generate_td_task_key,
)
from qcfractal.testing_helpers import run_service
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_service_client_error(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    client = snowflake.client()

    id_1, result_data_1 = submit_td_procedure_data(storage_socket, "td_H2O2_mopac_pm6", "test_tag", PriorityEnum.low)

    # Inject a failed computation
    failed_key = list(result_data_1.keys())[1]
    result_data_1[failed_key] = FailedOperation(
        error={"error_type": "test_error", "error_message": "this is just a test error"},
    )

    time_0 = now_at_utc()
    finished, n_optimizations = run_service(
        storage_socket, activated_manager_name, id_1, generate_td_task_key, result_data_1, 20
    )
    time_1 = now_at_utc()

    rec = client.get_torsiondrives(id_1)
    assert rec.status == RecordStatusEnum.error

    err_opts = []
    for optlist in rec.optimizations.values():
        err_opts.extend(opt for opt in optlist if opt.status == RecordStatusEnum.error)
    assert len(err_opts) == 1

    assert rec.children_status[RecordStatusEnum.error] == 1
    assert len(rec.children_errors) == 1

    assert rec.children_errors[0].status == RecordStatusEnum.error
    assert rec.children_errors[0].id == err_opts[0].id
    assert rec.children_errors[0].error["error_type"] == "test_error"
    assert rec.children_errors[0].error["error_message"] == "this is just a test error"



from __future__ import annotations

import contextlib
import copy
import importlib
import io
import json
import logging
from typing import TYPE_CHECKING

import sqlalchemy.orm.attributes

from ...db_socket.helpers import insert_general

try:
    from pydantic.v1 import BaseModel, Extra
except ImportError:
    from pydantic import BaseModel, Extra
from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import array_agg, aggregate_order_by
from sqlalchemy.orm import lazyload, selectinload, joinedload, defer, undefer

from qcfractal.components.optimization.record_db_models import (
    OptimizationSpecificationORM,
)
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcportal.exceptions import MissingDataError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.torsiondrive import (
    serialize_key,
    TorsiondriveSpecification,
    TorsiondriveQueryFilters,
    TorsiondriveInput,
    TorsiondriveMultiInput,
    TorsiondriveRecord,
)
from qcportal.utils import hash_dict, is_included
from .record_db_models import (
    TorsiondriveSpecificationORM,
    TorsiondriveInitialMoleculeORM,
    TorsiondriveOptimizationORM,
    TorsiondriveRecordORM,
)
from ..base_record_socket import BaseRecordSocket
from ..record_utils import append_output

# Torsiondrive package is optional
_td_spec = importlib.util.find_spec("torsiondrive")

if _td_spec is not None:
    _td_api_spec = importlib.util.find_spec("torsiondrive.td_api")

    torsiondrive = importlib.util.module_from_spec(_td_spec)
    td_api = importlib.util.module_from_spec(_td_api_spec)

    _td_spec.loader.exec_module(torsiondrive)
    _td_api_spec.loader.exec_module(td_api)


if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Dict, Tuple, Optional, Sequence, Any, Union, Iterable


class TorsiondriveServiceState(BaseModel):
    """
    This represents the current state of a torsiondrive service
    """

    class Config:
        extra = Extra.forbid
        allow_mutation = True
        validate_assignment = True

    torsiondrive_state = {}

    # These are stored as JSON (ie, dict encoded into a string)
    # This makes for faster loads and makes them somewhat tamper-proof
    molecule_template: str
    dihedral_template: str


# Meaningless, but unique to torsiondrives
torsiondrive_insert_lock_id = 14200


class TorsiondriveRecordSocket(BaseRecordSocket):
    """
    Socket for handling torsiondrive computations
    """

    # Used by the base class
    record_orm = TorsiondriveRecordORM
    record_input_type = TorsiondriveInput
    record_multi_input_type = TorsiondriveMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def available(self) -> bool:
        return _td_spec is not None

    def initialize_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ):
        td_orm: TorsiondriveRecordORM = service_orm.record
        specification = TorsiondriveSpecification(**td_orm.specification.model_dict())
        initial_molecules: List[Dict[str, Any]] = [x.molecule.model_dict() for x in td_orm.initial_molecules]
        keywords = specification.keywords

        # Create a template from the first initial molecule
        # we will assume they all have the same symbols, etc
        # TODO - can simplify this after removing numpy from db (ie, just copy initial_molecules[0])
        molecule_template = Molecule(**initial_molecules[0]).dict(encoding="json")
        molecule_template.pop("id", None)
        molecule_template.pop("identifiers", None)

        # The torsiondrive package uses print, so capture that using contextlib
        # Also capture any warnings generated by that package
        logging.captureWarnings(True)
        td_stdout = io.StringIO()
        with contextlib.redirect_stdout(td_stdout):
            td_state = td_api.create_initial_state(
                dihedrals=keywords.dihedrals,
                grid_spacing=keywords.grid_spacing,
                elements=molecule_template["symbols"],
                init_coords=[x["geometry"].tolist() for x in initial_molecules],
                dihedral_ranges=keywords.dihedral_ranges,
                energy_decrease_thresh=keywords.energy_decrease_thresh,
                energy_upper_limit=keywords.energy_upper_limit,
            )

        logging.captureWarnings(False)
        stdout = td_stdout.getvalue()

        # Build dihedral template. Just for convenience later
        dihedral_template = []
        for idx in keywords.dihedrals:
            tmp = {"type": "dihedral", "indices": idx}
            dihedral_template.append(tmp)

        dihedral_template_str = json.dumps(dihedral_template)
        molecule_template_str = json.dumps(molecule_template)

        append_output(session, td_orm, OutputTypeEnum.stdout, stdout)

        service_state = TorsiondriveServiceState(
            torsiondrive_state=td_state,
            dihedral_template=dihedral_template_str,
            molecule_template=molecule_template_str,
        )

        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ) -> bool:
        td_orm: TorsiondriveRecordORM = service_orm.record

        # Always update with the current provenance
        td_orm.compute_history[-1].provenance = {
            "creator": "torsiondrive",
            "version": torsiondrive.__version__,
            "routine": "torsiondrive.td_api",
        }

        # Load the state from the service_state column
        service_state = TorsiondriveServiceState(**service_orm.service_state)

        # Sort by position
        # Fully sorting by the key is not important since that ends up being a key in the dict
        # All that matters is that position 1 for a particular key comes before position 2, etc
        complete_tasks = sorted(service_orm.dependencies, key=lambda x: x.extras["position"])

        # Populate task results needed by the torsiondrive package
        task_results = {}
        for task in complete_tasks:
            td_api_key = task.extras["td_api_key"]
            task_results.setdefault(td_api_key, [])

            # This is an ORM for an optimization
            opt_record = task.record

            # Lookup molecules
            initial_id = opt_record.initial_molecule_id
            final_id = opt_record.final_molecule_id
            mol_ids = [initial_id, final_id]
            mol_data = self.root_socket.molecules.get(molecule_id=mol_ids, include=["geometry"], session=session)

            # Use plain lists rather than numpy arrays
            initial_mol_geom = mol_data[0]["geometry"].tolist()
            final_mol_geom = mol_data[1]["geometry"].tolist()

            task_results[td_api_key].append((initial_mol_geom, final_mol_geom, opt_record.energies[-1]))

        # The torsiondrive package uses print, so capture that using contextlib
        # Also capture any warnings generated by that package
        td_stdout = io.StringIO()
        logging.captureWarnings(True)
        with contextlib.redirect_stdout(td_stdout):
            td_api.update_state(service_state.torsiondrive_state, task_results)
            next_tasks = td_api.next_jobs_from_state(service_state.torsiondrive_state, verbose=True)

        stdout_append = "\n" + td_stdout.getvalue()
        logging.captureWarnings(False)

        # If there are any tasks left, submit them
        if len(next_tasks) > 0:
            self._submit_optimizations(session, service_state, service_orm, next_tasks)
        else:
            # check that what we have is consistent with what the torsiondrive package reports
            lowest_energies = td_api.collect_lowest_energies(service_state.torsiondrive_state)
            lowest_energies = {serialize_key(x): y for x, y in lowest_energies.items()}

            our_energies = {x.key: [] for x in td_orm.optimizations}
            for x in td_orm.optimizations:
                if x.energy is not None:
                    our_energies[x.key].append(x.energy)

            min_energies = {x: min(y) if y else None for x, y in our_energies.items()}
            if lowest_energies != min_energies:
                raise RuntimeError("Minimum energies reported by the torsiondrive package do not match ours!")

        # append to the existing stdout
        append_output(session, td_orm, OutputTypeEnum.stdout, stdout_append)

        # Set the new service state. We must then mark it as modified
        # so that SQLAlchemy can pick up changes. This is because SQLAlchemy
        # cannot track mutations in nested dicts
        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")

        # Return True to indicate that this service has successfully completed
        return len(next_tasks) == 0

    def _submit_optimizations(
        self,
        session: Session,
        service_state: TorsiondriveServiceState,
        service_orm: ServiceQueueORM,
        next_tasks: Dict[str, Any],
    ):
        """
        Submit the next batch of optimizations for a torsiondrive service
        """

        td_orm: TorsiondriveRecordORM = service_orm.record

        # delete all existing entries in the dependency list
        service_orm.dependencies = []

        # Create an optimization input based on the new geometry and the optimization template
        opt_spec = td_orm.specification.optimization_specification.model_dict()

        # Convert to an input
        opt_spec = OptimizationSpecification(**opt_spec).dict()

        for td_api_key, geometries in next_tasks.items():
            # Make a deep copy to prevent modifying the original ORM
            opt_spec2 = copy.deepcopy(opt_spec)

            # Construct constraints
            constraints = json.loads(service_state.dihedral_template)

            grid_id = td_api.grid_id_from_string(td_api_key)
            for con_num, k in enumerate(grid_id):
                constraints[con_num]["value"] = k

            # update the constraints
            opt_spec2["keywords"].setdefault("constraints", {})
            opt_spec2["keywords"]["constraints"].setdefault("set", [])
            opt_spec2["keywords"]["constraints"]["set"].extend(constraints)

            # Loop over the new geometries from the torsiondrive package
            constrained_mols = []
            for geometry in geometries:
                # Build new molecule
                mol = json.loads(service_state.molecule_template)
                mol["geometry"] = geometry

                constrained_mols.append(Molecule(**mol))

            # Submit the new optimizations
            meta, opt_ids = self.root_socket.records.optimization.add(
                constrained_mols,
                OptimizationSpecification(**opt_spec2),
                service_orm.compute_tag,
                service_orm.compute_priority,
                td_orm.creator_user_id,
                service_orm.find_existing,
                session=session,
            )

            if not meta.success:
                raise RuntimeError("Error adding optimizations - likely a developer error: " + meta.error_string)

            # ids will be in the same order as the molecules (and the geometries from td)
            opt_key = serialize_key(grid_id)
            for position, opt_id in enumerate(opt_ids):
                svc_dep = ServiceDependencyORM(
                    record_id=opt_id,
                    extras={"td_api_key": td_api_key, "position": position},
                )

                # The position field is handled by the collection class in sqlalchemy
                # corresponds to the absolute position across all optimizations for this torsiondrive,
                # not the position of the geometry for this td_api_key (as stored in the ServiceDependenciesORM)
                opt_history = TorsiondriveOptimizationORM(
                    torsiondrive_id=service_orm.record_id,
                    optimization_id=opt_id,
                    key=opt_key,
                )

                service_orm.dependencies.append(svc_dep)
                td_orm.optimizations.append(opt_history)

    def insert_full_qcportal_records_v1(
        self,
        session: Session,
        records: Sequence[TorsiondriveRecord],
        creator_user_id: Optional[int],
    ) -> List[TorsiondriveRecordORM]:
        ret = []

        for record in records:
            meta, spec_ids = self.root_socket.records.torsiondrive.add_specifications(
                [record.specification], session=session
            )
            if not meta.success:
                raise RuntimeError(
                    "Aborted torsiondrive insertion - could not add specifications: " + meta.error_string
                )

            meta, initial_mol_ids = self.root_socket.molecules.add(record.initial_molecules, session=session)
            if not meta.success:
                raise RuntimeError(
                    "Aborted torsiondrive insertion - could not add initial molecules: " + meta.error_string
                )

            init_mol_orms = [TorsiondriveInitialMoleculeORM(molecule_id=m) for m in initial_mol_ids]

            record_orm = TorsiondriveRecordORM(
                specification_id=spec_ids[0],
                initial_molecules=init_mol_orms,
                status=RecordStatusEnum.complete,
            )

            if record.optimizations:
                # a map of original record ids to the td opt info
                td_opt_map = {o.optimization_id: o for o in record.optimizations_}

                record_orm.optimizations = []
                for k, opt_records in record.optimizations.items():
                    # Just to be safe
                    assert all(serialize_key(k) == td_opt_map[r.id].key for r in opt_records)

                    opt_ids = self.root_socket.records.insert_full_qcportal_records(
                        session, opt_records, creator_user_id
                    )
                    opt_orm = [
                        TorsiondriveOptimizationORM(
                            optimization_id=oid, key=td_opt_map[r.id].key, position=td_opt_map[r.id].position
                        )
                        for r, oid in zip(opt_records, opt_ids)
                    ]

                    record_orm.optimizations.extend(opt_orm)

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, td_specs: Sequence[TorsiondriveSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specifications for torsiondrive services to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        td_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for td_spec in td_specs:
            td_kw_dict = td_spec.keywords.dict()

            td_spec_dict = {"program": td_spec.program, "keywords": td_kw_dict, "protocols": {}}
            td_spec_hash = hash_dict(td_spec_dict)

            td_spec_orm = TorsiondriveSpecificationORM(
                program=td_spec.program,
                keywords=td_kw_dict,
                protocols=td_spec_dict["protocols"],
                specification_hash=td_spec_hash,
            )

            to_add.append(td_spec_orm)

        with self.root_socket.optional_session(session, False) as session:

            opt_specs = [x.optimization_specification for x in td_specs]
            meta, opt_spec_ids = self.root_socket.records.optimization.add_specifications(opt_specs, session=session)

            if not meta.success:
                return (
                    InsertMetadata(
                        error_description="Unable to add optimization specifications: " + meta.error_string,
                    ),
                    [],
                )

            assert len(opt_spec_ids) == len(td_specs)
            for td_spec_orm, opt_spec_id in zip(to_add, opt_spec_ids):
                td_spec_orm.optimization_specification_id = opt_spec_id

            meta, ids = insert_general(
                session,
                to_add,
                (
                    TorsiondriveSpecificationORM.specification_hash,
                    TorsiondriveSpecificationORM.optimization_specification_id,
                ),
                (TorsiondriveSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, td_spec: TorsiondriveSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for a torsiondrive service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        td_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([td_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []
        if include:
            # Initial molecules will get both the ids and the actual molecule
            if is_included("initial_molecules", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(
                    selectinload(TorsiondriveRecordORM.initial_molecules).joinedload(
                        TorsiondriveInitialMoleculeORM.molecule
                    )
                )
            elif is_included("initial_molecules_ids", include, exclude, False):
                options.append(selectinload(TorsiondriveRecordORM.initial_molecules))

            if is_included("optimizations", include, exclude, False):
                options.append(selectinload(TorsiondriveRecordORM.optimizations))

            if is_included("minimum_optimizations", include, exclude, False):
                options.append(undefer(TorsiondriveRecordORM.minimum_optimizations))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: TorsiondriveQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_spspec_join = False
        need_optspec_join = False
        need_initmol_join = False

        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_spspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_spspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_spspec_join = True
        if query_data.optimization_program is not None:
            and_query.append(OptimizationSpecificationORM.program.in_(query_data.optimization_program))
            need_optspec_join = True
        if query_data.initial_molecule_id is not None:
            and_query.append(TorsiondriveInitialMoleculeORM.molecule_id.in_(query_data.initial_molecule_id))
            need_initmol_join = True

        stmt = select(TorsiondriveRecordORM.id)

        # We don't search for anything td-specification specific, so no need for
        # need_tdspec_join (for now...)

        if need_optspec_join or need_spspec_join:
            stmt = stmt.join(TorsiondriveRecordORM.specification)
            stmt = stmt.join(TorsiondriveSpecificationORM.optimization_specification)

        if need_spspec_join:
            stmt = stmt.join(OptimizationSpecificationORM.qc_specification)

        if need_initmol_join:
            stmt = stmt.join(
                TorsiondriveInitialMoleculeORM,
                TorsiondriveInitialMoleculeORM.torsiondrive_id == TorsiondriveRecordORM.id,
            )

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=TorsiondriveRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_molecule_ids: Sequence[Iterable[int]],
        td_spec_id: int,
        as_service: bool,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new torsiondrive computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecule_ids
            IDs of the initial sets of molecules to start the torsiondrive. One record will be added per set.
        td_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who creasted the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            # Lock for the entire transaction
            session.execute(select(func.pg_advisory_xact_lock(torsiondrive_insert_lock_id))).scalar()

            td_ids = []
            inserted_idx = []
            existing_idx = []

            if find_existing:
                # Torsiondrives are a bit more complicated because we have a many-to-many relationship
                # between torsiondrives and initial molecules. So skip the general insert
                # function and do this one at a time

                # Create a cte with the initial molecules we can query against
                # This is like a table, with the specification id and the initial molecule ids
                # as a postgres array (sorted)
                # We then use this to determine if there are duplicates
                init_mol_cte = (
                    select(
                        TorsiondriveRecordORM.id,
                        TorsiondriveRecordORM.specification_id,
                        array_agg(
                            aggregate_order_by(
                                TorsiondriveInitialMoleculeORM.molecule_id,
                                TorsiondriveInitialMoleculeORM.molecule_id.asc(),
                            )
                        ).label("molecule_ids"),
                    )
                    .join(
                        TorsiondriveInitialMoleculeORM,
                        TorsiondriveInitialMoleculeORM.torsiondrive_id == TorsiondriveRecordORM.id,
                    )
                    .group_by(TorsiondriveRecordORM.id)
                    .cte()
                )

                for idx, mol_ids in enumerate(initial_molecule_ids):
                    # sort molecules by increasing ids, and remove duplicates
                    mol_ids = sorted(set(mol_ids))

                    # does this exist?
                    stmt = select(init_mol_cte.c.id)
                    stmt = stmt.where(init_mol_cte.c.specification_id == td_spec_id)
                    stmt = stmt.where(init_mol_cte.c.molecule_ids == mol_ids)
                    existing = session.execute(stmt).scalars().first()

                    if not existing:
                        td_orm = TorsiondriveRecordORM(
                            is_service=as_service,
                            specification_id=td_spec_id,
                            status=RecordStatusEnum.waiting,
                            creator_user_id=creator_user_id,
                        )

                        self.create_service(td_orm, compute_tag, compute_priority, find_existing)

                        session.add(td_orm)
                        session.flush()

                        for mid in mol_ids:
                            mid_orm = TorsiondriveInitialMoleculeORM(molecule_id=mid, torsiondrive_id=td_orm.id)
                            session.add(mid_orm)

                        session.flush()

                        td_ids.append(td_orm.id)
                        inserted_idx.append(idx)
                    else:
                        td_ids.append(existing)
                        existing_idx.append(idx)
            else:  # not finding existing - just add all
                for idx, mol_ids in enumerate(initial_molecule_ids):
                    # sort molecules by increasing ids, and remove duplicates
                    mol_ids = sorted(set(mol_ids))

                    td_orm = TorsiondriveRecordORM(
                        is_service=as_service,
                        specification_id=td_spec_id,
                        status=RecordStatusEnum.waiting,
                        creator_user_id=creator_user_id,
                    )

                    self.create_service(td_orm, compute_tag, compute_priority, find_existing)

                    session.add(td_orm)
                    session.flush()

                    for mid in mol_ids:
                        mid_orm = TorsiondriveInitialMoleculeORM(molecule_id=mid, torsiondrive_id=td_orm.id)
                        session.add(mid_orm)

                    session.flush()

                    td_ids.append(td_orm.id)
                    inserted_idx.append(idx)

            meta = InsertMetadata(inserted_idx=inserted_idx, existing_idx=existing_idx)
            return meta, td_ids

    def add(
        self,
        initial_molecules: Sequence[Iterable[Union[int, Molecule]]],
        td_spec: TorsiondriveSpecification,
        as_service: bool,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[str],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new torsiondrive calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecules
            Initial sets of molecules to start the torsiondrive. One record will be added per set.
        td_spec
            Specification for the calculations
        as_service
            Whether this record should be run as a service or as a regular calculation
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Nameor ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(td_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            init_mol_ids = []
            for init_mol in initial_molecules:
                mol_meta, mol_ids = self.root_socket.molecules.add_mixed(init_mol, session=session)
                if not mol_meta.success:
                    return (
                        InsertMetadata(
                            error_description="Aborted - could not add all molecules: " + mol_meta.error_string
                        ),
                        [],
                    )

                init_mol_ids.append(mol_ids)

            return self.add_internal(
                init_mol_ids,
                spec_id,
                as_service,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: TorsiondriveInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, TorsiondriveInput)

        meta, ids = self.add(
            [record_input.initial_molecules],
            record_input.specification,
            True,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for torsiondrives
    ####################################################

    def get_initial_molecules_ids(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Obtain the initial molecules of a torsiondrive

        Parameters
        ----------
        record_id
            ID of the torsiondrive to get the minimum optimizations of
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of Molecule ids
        """

        options = [
            lazyload("*"),
            defer("*"),
            joinedload(TorsiondriveRecordORM.initial_molecules).options(
                undefer(TorsiondriveInitialMoleculeORM.molecule_id)
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(TorsiondriveRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.molecule_id for x in rec.initial_molecules]

    def get_optimizations(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(TorsiondriveRecordORM.optimizations).options(undefer("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(TorsiondriveRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.optimizations]

    def get_minimum_optimizations(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, int]:
        """
        Obtain the records for the minimum optimizations for a torsiondrive

        Parameters
        ----------
        record_id
            ID of the torsiondrive to get the minimum optimizations of
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dictionary, where the value is an optimization record (as a dictionary) and the key is the key of the
            optimization in the torsiondrive (representing the angles)
        """

        stmt = select(TorsiondriveRecordORM.minimum_optimizations).where(TorsiondriveRecordORM.id == record_id)

        with self.root_socket.optional_session(session, True) as session:
            r = session.execute(stmt).scalar_one_or_none()  # List of (key, id)
            return {} if r is None else r



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.optimization.record_db_models import OptimizationRecordORM
from qcfractal.components.optimization.testing_helpers import (
    test_specs,
    load_procedure_data,
    run_procedure_data,
    load_record_data,
)
from qcfractal.components.testing_helpers import convert_to_plain_qcschema_result
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.managers import ManagerName
from qcportal.molecules import Molecule
from qcportal.optimization import (
    compare_optimization_records,
    OptimizationSpecification,
)
from qcportal.record_models import RecordStatusEnum, PriorityEnum, RecordTask
from qcportal.singlepoint import (
    QCSpecification,
    SinglepointDriver,
    SinglepointProtocols,
)
from qcportal.utils import now_at_utc
from ..record_utils import build_extras_properties

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from sqlalchemy.orm.session import Session
    from typing import List, Dict


def _compare_record_with_schema(record_orm, result_schema):
    assert record_orm.status == RecordStatusEnum.complete
    assert record_orm.specification.program == result_schema.provenance.creator.lower()

    kw_no_prog = result_schema.keywords.copy()
    kw_no_prog["program"] = result_schema.keywords["program"]
    assert kw_no_prog == result_schema.keywords

    # The singlepoint spec
    assert record_orm.specification.qc_specification.program == result_schema.keywords["program"]
    assert record_orm.specification.qc_specification.method == result_schema.input_specification.model.method
    assert record_orm.specification.qc_specification.basis == result_schema.input_specification.model.basis
    assert record_orm.specification.qc_specification.keywords == result_schema.input_specification.keywords

    assert len(record_orm.compute_history) == 1
    assert record_orm.compute_history[0].status == RecordStatusEnum.complete
    assert record_orm.compute_history[0].provenance == result_schema.provenance

    # Test the trajectory
    assert len(record_orm.trajectory) == len(result_schema.trajectory)
    for db_traj, res_traj in zip(record_orm.trajectory, result_schema.trajectory):
        assert db_traj.singlepoint_record.specification.program == res_traj.provenance.creator.lower()
        assert db_traj.singlepoint_record.specification.basis == res_traj.model.basis
        assert db_traj.singlepoint_record.molecule.identifiers["molecule_hash"] == res_traj.molecule.get_hash()

    # Use plain schema, where compressed stuff is removed
    new_extras, new_properties = build_extras_properties(result_schema.copy(deep=True))
    assert record_orm.properties == new_properties
    assert record_orm.extras == new_extras

    # TODO - eventually schema may have these
    assert record_orm.native_files == {}

    for k in ("stdout", "stderr", "error"):
        plain_output = getattr(result_schema, k)
        if plain_output is not None:
            out_str = record_orm.compute_history[0].outputs[k].get_output()
            assert out_str == plain_output
        else:
            assert k not in record_orm.compute_history[0].outputs


@pytest.mark.parametrize("spec", test_specs)
def test_optimization_socket_task_spec(
    storage_socket: SQLAlchemySocket,
    spec: OptimizationSpecification,
    activated_manager_name: ManagerName,
    activated_manager_programs: Dict[str, List[str]],
):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    time_0 = now_at_utc()
    meta, id = storage_socket.records.optimization.add(all_mols, spec, "tag1", PriorityEnum.low, None, True)
    time_1 = now_at_utc()
    assert meta.success

    tasks = storage_socket.tasks.claim_tasks(activated_manager_name.fullname, activated_manager_programs, ["*"])
    tasks = [RecordTask(**t) for t in tasks]

    assert len(tasks) == 3
    for t in tasks:
        assert t.function == "qcengine.compute_procedure"
        assert t.function_kwargs["procedure"] == spec.program

        kw_with_prog = spec.keywords.copy()
        kw_with_prog["program"] = spec.qc_specification.program

        task_input = t.function_kwargs["input_data"]
        assert task_input["keywords"] == kw_with_prog
        assert task_input["protocols"] == spec.protocols.dict(exclude_defaults=True, exclude_unset=True)

        # Forced to gradient in the qcschema input
        assert task_input["input_specification"]["driver"] == SinglepointDriver.gradient
        assert task_input["input_specification"]["model"] == {
            "method": spec.qc_specification.method,
            "basis": spec.qc_specification.basis,
        }

        assert task_input["input_specification"]["keywords"] == spec.qc_specification.keywords

        assert t.compute_tag == "tag1"
        assert t.compute_priority == PriorityEnum.low

    rec_id_mol_map = {
        id[0]: all_mols[0],
        id[1]: all_mols[1],
        id[2]: all_mols[2],
    }

    assert Molecule(**tasks[0].function_kwargs["input_data"]["initial_molecule"]) == rec_id_mol_map[tasks[0].record_id]
    assert Molecule(**tasks[1].function_kwargs["input_data"]["initial_molecule"]) == rec_id_mol_map[tasks[1].record_id]
    assert Molecule(**tasks[2].function_kwargs["input_data"]["initial_molecule"]) == rec_id_mol_map[tasks[2].record_id]


def test_optimization_socket_find_existing_1(storage_socket: SQLAlchemySocket):
    spec = OptimizationSpecification(
        program="optprog1",
        keywords={},
        protocols={"trajectory": "initial_and_final"},
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.optimization.add([water], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.optimization.add([water], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_optimization_socket_find_existing_2(storage_socket: SQLAlchemySocket):
    # Test case sensitivity
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={},
        protocols={"trajectory": "initial_and_final"},
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    spec2 = OptimizationSpecification(
        program="opTPROg1",
        keywords={},
        protocols={"trajectory": "initial_and_final"},
        qc_specification=QCSpecification(
            program="prOG1",
            driver="deferred",
            method="b3LYp",
            basis="6-31g*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.optimization.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.optimization.add([water], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_optimization_socket_find_existing_3(storage_socket: SQLAlchemySocket):
    # Test default keywords and protocols
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={},
        protocols={},
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    spec2 = OptimizationSpecification(
        program="optprog1",
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "value"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.optimization.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.optimization.add([water], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_optimization_socket_find_existing_4(storage_socket: SQLAlchemySocket):
    # Test adding molecule by id

    water = load_molecule_data("water_dimer_minima")
    kw = {"a": "value"}
    _, mol_ids = storage_socket.molecules.add([water])

    spec1 = OptimizationSpecification(
        program="optprog1",
        qc_specification=QCSpecification(
            program="prog1",
            driver="deferred",
            method="b3lyp",
            basis="6-31G*",
        ),
    )

    meta, id1 = storage_socket.records.optimization.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.optimization.add(mol_ids, spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_optimization_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName
):
    test_names = [
        "opt_psi4_fluoroethane_notraj",
        "opt_psi4_benzene",
        "opt_psi4_methane_sometraj",
    ]

    all_results = []
    all_id = []

    for test_name in test_names:
        _, _, result_data = load_procedure_data(test_name)
        record_id = run_procedure_data(storage_socket, activated_manager_name, test_name)
        all_results.append(result_data)
        all_id.append(record_id)

    for rec_id, result in zip(all_id, all_results):
        record = session.get(OptimizationRecordORM, rec_id)

        plain_result = convert_to_plain_qcschema_result(result)
        _compare_record_with_schema(record, plain_result)


def test_optimization_socket_insert_full_schema_v1(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "opt_psi4_benzene",
        "opt_psi4_fluoroethane_notraj",
        "opt_psi4_methane",
        "opt_psi4_methane_sometraj",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    all_ids = []

    for test_name in test_names:
        _, _, result_schema = load_procedure_data(test_name)

        plain_schema = convert_to_plain_qcschema_result(result_schema)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session2:
            ins_ids_1 = storage_socket.records.insert_full_schema_v1(session2, [result_schema.copy(deep=True)], user_id)
            ins_ids_2 = storage_socket.records.insert_full_schema_v1(session2, [plain_schema.copy(deep=True)], user_id)

        ins_id_1 = ins_ids_1[0]
        ins_id_2 = ins_ids_2[0]

        # insert_full_schema always inserts
        assert ins_id_1 != ins_id_2
        assert ins_id_1 not in all_ids
        assert ins_id_2 not in all_ids
        all_ids.extend([ins_id_1, ins_id_2])

        with storage_socket.session_scope() as session:
            rec_1 = session.get(OptimizationRecordORM, ins_id_1)
            rec_2 = session.get(OptimizationRecordORM, ins_id_2)

            assert rec_1.creator_user_id == user_id
            assert rec_2.creator_user_id == user_id

            _compare_record_with_schema(rec_1, plain_schema)
            _compare_record_with_schema(rec_2, plain_schema)


def test_optimization_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "opt_psi4_benzene",
        "opt_psi4_fluoroethane_notraj",
        "opt_psi4_methane",
        "opt_psi4_methane_sometraj",
        "opt_error_118868739",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_optimizations(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)

        compare_optimization_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"
        for sp in rec_1.trajectory:
            assert sp.creator_user == "submit_user"



"""Fix task queue sort

Revision ID: 8263992eb6c8
Revises: f8a7c273f18a
Create Date: 2024-12-18 11:34:57.186942

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "8263992eb6c8"
down_revision = "f8a7c273f18a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.execute(
        "CREATE INDEX ix_task_queue_sort ON task_queue (priority DESC, sort_date, id, tag) WHERE available = True;"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    String,
    ForeignKey,
    Index,
    LargeBinary,
    CheckConstraint,
    UniqueConstraint,
    Boolean,
)
from sqlalchemy.dialects.postgresql import ARRAY, TEXT, TIMESTAMP
from sqlalchemy.orm import relationship

from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import BaseORM
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class TaskQueueORM(BaseORM):
    """
    Table for storing information about tasks
    """

    __tablename__ = "task_queue"

    id = Column(Integer, primary_key=True)

    function = Column(String, nullable=True)
    function_kwargs_compressed = Column(LargeBinary, nullable=True)

    # For some reason, this can't be array of varchar. If it is, the comparisons
    # when claiming tasks don't work
    required_programs = Column(ARRAY(TEXT), nullable=False)

    sort_date = Column(TIMESTAMP(timezone=True), default=now_at_utc, nullable=False)
    compute_tag = Column(String, nullable=False)
    compute_priority = Column(Integer, nullable=False)
    available = Column(Boolean, nullable=False)

    record_id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), nullable=False)
    record = relationship(BaseRecordORM, back_populates="task", uselist=False)

    # An important special case is ORDER BY in combination with LIMIT n: an
    # explicit sort will have to process all the data to identify the first n
    # rows, but if there is an index matching the ORDER BY, the first n rows
    # can be retrieved directly, without scanning the remainder at all.
    __table_args__ = (
        Index("ix_task_queue_tag", "compute_tag"),
        Index("ix_task_queue_required_programs", "required_programs", postgresql_using="gin"),
        Index(
            "ix_task_queue_sort",
            compute_priority.desc(),
            sort_date.asc(),
            id.asc(),
            compute_tag,
            postgresql_where=(available == True),
        ),
        UniqueConstraint("record_id", name="ux_task_queue_record_id"),
        # WARNING - these are not autodetected by alembic
        CheckConstraint(
            "required_programs::text = LOWER(required_programs::text)", name="ck_task_queue_requirements_lower"
        ),
        CheckConstraint("compute_tag = LOWER(compute_tag)", name="ck_task_queue_compute_tag_lower"),
    )

    # Remove sort_date from the model. For backwards compatibility (and because it's only used for sorting)
    # Also remove the "available" column - is somewhat redundant with the record status
    _qcportal_model_excludes = ["sort_date", "available"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - DEPRECATED - remove eventually
        if "compute_tag" in d:
            d["tag"] = d.pop("compute_tag")
        if "compute_priority" in d:
            d["priority"] = d.pop("compute_priority")

        return d



"""add user id to session table

Revision ID: 1136ef3ba3ef
Revises: 521464fa93ca
Create Date: 2025-04-05 09:47:58.918236

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1136ef3ba3ef"
down_revision = "521464fa93ca"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("DELETE FROM user_session")
    op.add_column("user_session", sa.Column("user_id", sa.Integer(), nullable=False))
    op.create_index("ix_user_session_user_id", "user_session", ["user_id"], unique=False)
    op.create_foreign_key(None, "user_session", "user", ["user_id"], ["id"], ondelete="cascade")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "user_session", type_="foreignkey")
    op.drop_index("ix_user_session_user_id", table_name="user_session")
    op.drop_column("user_session", "user_id")
    # ### end Alembic commands ###
